"""Constraints that wrap or contain other constraints"""

from __future__ import annotations

from pathlib import Path
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
)

from datalad_next.exceptions import CapturedException

from .base import (
    Constraint,
    ConstraintError,
    DatasetParameter,
)


class EnsureIterableOf(Constraint):
    """Ensure that an input is a list of a particular data type
    """
    # TODO support a delimiter to be able to take str-lists?
    def __init__(self,
                 iter_type: type,
                 item_constraint: Callable,
                 min_len: int | None = None,
                 max_len: int | None = None):
        """
        Parameters
        ----------
        iter_type:
          Target type of iterable. Common types are `list`, or `tuple`,
          but also generator type iterables are possible. Type constructor
          must take an iterable with items as the only required positional
          argument.
        item_constraint:
          Each incoming item will be mapped through this callable
          before being passed to the iterable type constructor.
        min_len:
          If not None, the iterable will be verified to have this minimum
          number of items. The iterable type must implement `__len__()`
          for this check to be supported.
        max_len:
          If not None, the iterable will be verified to have this maximum
          number of items. The iterable type must implement `__len__()`
          for this check to be supported.
        """
        if min_len is not None and max_len is not None and min_len > max_len:
            raise ValueError(
                'Given minimum length exceeds given maximum length')
        self._iter_type = iter_type
        self._item_constraint = item_constraint
        self._min_len = min_len
        self._max_len = max_len
        super().__init__()

    def __repr__(self):
        # not showing iter_type here, will come via class.name
        # in general
        return (
            f'{self.__class__.__name__}('
            f'item_constraint={self._item_constraint!r}'
            f', min_len={self._min_len!r}'
            f', max_len={self._max_len!r})'
        )

    @property
    def item_constraint(self):
        return self._item_constraint

    def __call__(self, value):
        try:
            iter = self._iter_type(
                self._item_constraint(i) for i in value
            )
        except (ConstraintError, TypeError) as e:
            self.raise_for(
                value,
                "{itertype} item is not {itype}",
                itertype=self._iter_type.__name__,
                itype=self._item_constraint,
                __caused_by__=e,
            )
        if self._min_len is not None or self._max_len is not None:
            # only do this if necessary, generators will not support
            # __len__, for example
            iter_len = len(iter)
            if self._min_len is not None and iter_len < self._min_len:
                self.raise_for(
                    iter,
                    'must have minimum length {len}',
                    len=self._min_len,
                )
            if self._max_len is not None and iter_len > self._max_len:
                self.raise_for(
                    iter,
                    'must not exceed maximum length {len}',
                    len=self._max_len,
                )
        return iter

    def short_description(self):
        return f'{self._iter_type}({self._item_constraint})'


class EnsureListOf(EnsureIterableOf):
    def __init__(self,
                 item_constraint: Callable,
                 min_len: int | None = None,
                 max_len: int | None = None):
        """
        Parameters
        ----------
        item_constraint:
          Each incoming item will be mapped through this callable
          before being passed to the list constructor.
        min_len:
          If not None, the list will be verified to have this minimum
          number of items.
        max_len:
          If not None, the list will be verified to have this maximum
          number of items.
        """
        super().__init__(list, item_constraint,
                         min_len=min_len, max_len=max_len)

    def short_description(self):
        return f'list({self._item_constraint})'


class EnsureTupleOf(EnsureIterableOf):
    def __init__(self,
                 item_constraint: Callable,
                 min_len: int | None = None,
                 max_len: int | None = None):
        """
        Parameters
        ----------
        item_constraint:
          Each incoming item will be mapped through this callable
          before being passed to the tuple constructor.
        min_len:
          If not None, the tuple will be verified to have this minimum
          number of items.
        max_len:
          If not None, the tuple will be verified to have this maximum
          number of items.
        """
        super().__init__(tuple, item_constraint,
                         min_len=min_len, max_len=max_len)

    def short_description(self):
        return f'tuple({self._item_constraint})'


class EnsureMapping(Constraint):
    """Ensure a mapping of a key to a value of a specific nature"""

    def __init__(self,
                 key: Constraint,
                 value: Constraint,
                 delimiter: str = ':',
                 allow_length2_sequence: bool = True):
        """
        Parameters
        ----------
        key:
          Key constraint instance.
        value:
          Value constraint instance.
        delimiter:
          Delimiter to use for splitting a key from a value for a `str` input.
        """
        super().__init__()
        self._key_constraint = key
        self._value_constraint = value
        self._delimiter = delimiter
        self._allow_length2_sequence = allow_length2_sequence

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'key={self._key_constraint!r}'
            f', value={self._value_constraint!r}'
            f', delimiter={self._delimiter!r})'
        )

    def short_description(self):
        return 'mapping of {} -> {}'.format(
            self._key_constraint.short_description(),
            self._value_constraint.short_description(),
        )

    def _get_key_value(self, value) -> tuple:
        # determine key and value from various kinds of input
        if isinstance(value, str):
            # will raise if it cannot split into two
            key, val = value.split(sep=self._delimiter, maxsplit=1)
        elif isinstance(value, dict):
            if not len(value):
                self.raise_for(value, 'dict does not contain a key')
            elif len(value) > 1:
                self.raise_for(value, 'dict contains more than one key')
            key, val = value.copy().popitem()
        elif self._allow_length2_sequence and isinstance(value, (list, tuple)):
            if not len(value) == 2:
                self.raise_for(value, 'key/value sequence does not have length 2')
            key, val = value
        else:
            self.raise_for(value, 'not a recognized mapping')

        return key, val

    def __call__(self, value) -> Dict:
        key, val = self._get_key_value(value)
        key = self._key_constraint(key)
        val = self._value_constraint(val)
        return {key: val}

    def for_dataset(self, dataset: DatasetParameter) -> Constraint:
        # tailor both constraints to the dataset and reuse delimiter
        return EnsureMapping(
            key=self._key_constraint.for_dataset(dataset),
            value=self._value_constraint.for_dataset(dataset),
            delimiter=self._delimiter,
        )


class EnsureGeneratorFromFileLike(Constraint):
    """Ensure a constraint for each item read from a file-like.

    A given value can either be a file-like (the outcome of `open()`,
    or `StringIO`), or `-` as an alias of STDIN, or a path to an
    existing file to be read from.
    """

    def __init__(
        self,
        item_constraint: Callable,
        exc_mode: str = 'raise',
    ):
        """
        Parameters
        ----------
        item_constraint:
          Each incoming item will be mapped through this callable
          before being yielded by the generator.
        exc_mode: {'raise', 'yield'}, optional
          How to deal with exceptions occurring when processing
          individual lines/items. With 'yield' the respective
          exception instance is yielded as a ``CapturedException``,
          and processing continues.
          A caller can then decide whether to ignore, or report the
          exception. With 'raise', an exception is raised immediately
          and processing stops.
        """
        assert exc_mode in ('raise', 'yield')
        self._item_constraint = item_constraint
        self._exc_mode = exc_mode
        super().__init__()

    def __repr__(self):
        # not showing iter_type here, will come via class.name
        # in general
        return (
            f'{self.__class__.__name__}('
            f'item_constraint={self._item_constraint!r})'
        )

    def short_description(self):
        return \
            f'items of type "{self._item_constraint.short_description()}" ' \
            'read from a file-like'

    def __call__(self, value) -> Generator[Any, None, None]:
        # we only support a single file-like source. If we happened to get
        # a length-1 sequence (for technical reasons, such as argparse
        # having collected the value), we unpack it.
        if isinstance(value, (list, tuple)) and len(value) == 1:
            value = value[0]
        opened_file = False
        if value == '-':
            value = sys.stdin
        elif isinstance(value, (str, Path)):
            # we covered the '-' special case, so this must be a Path
            path = Path(value) if not isinstance(value, Path) else value
            if not path.is_file():
                self.raise_for(
                    value,
                    "not '-', or a path to an existing file",
                )
            value = path.open()
            opened_file = True
        return self._item_yielder(value, opened_file)

    def _item_yielder(self, fp, close_file):
        try:
            for line in fp:
                try:
                    yield self._item_constraint(
                        # splitlines() removes the newline at the end of
                        # the string that is left in by __iter__()
                        line.splitlines()[0]
                    )
                except Exception as e:
                    if self._exc_mode == 'raise':
                        raise
                    else:
                        yield CapturedException(e)
        finally:
            if close_file:
                fp.close()


class ConstraintWithPassthrough(Constraint):
    """Regular constraint, but with a "pass-through" value that is not processed

    This is different from a `Constraint() | EnsureValue(...)` construct,
    because the pass-through value is not communicated. This can be useful
    when a particular value must be supported for technical reasons, but
    need not, or must not be included in (error) messages.

    The pass-through is returned as-is, and is not processed except for an
    identity check (`==`).

    For almost all reporting (`__str__`, descriptions, ...) the wrapped
    value constraint is used, making this class virtually invisible.
    Only ``__repr__`` reflects the wrapping.
    """
    def __init__(self,
                 constraint: Constraint,
                 passthrough: Any):
        """
        Parameters
        ----------
        constraint: Constraint
          Any ``Constraint`` subclass instance that will be used to validate
          values.
        passthrough:
          A value that will not be subjected to validation by the value
          constraint, but is returned as-is. This value is not copied.
          It is a caller's responsibility to guarantee immutability if that
          is desired.
        """
        super().__init__()
        self._constraint = constraint
        self._passthrough = passthrough

    @property
    def constraint(self) -> Constraint:
        """Returns the wrapped constraint instance"""
        return self._constraint

    @property
    def passthrough(self) -> Any:
        """Returns the set pass-through value"""
        return self._passthrough

    def __call__(self, value) -> Any:
        if value == self._passthrough:
            val = value
        else:
            val = self._constraint(value)
        return val

    def __str__(self) -> str:
        return self._constraint.__str__()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}' \
               f'({self._constraint!r}, passthrough={self._passthrough!r})'

    def for_dataset(self, dataset: DatasetParameter) -> Constraint:
        """Wrap the wrapped constraint again after tailoring it for the dataset

        The pass-through value is reused.
        """
        return self.__class__(
            self._constraint.for_dataset(dataset),
            passthrough=self._passthrough,
        )

    def long_description(self) -> str:
        return self._constraint.long_description()

    def short_description(self) -> str:
        return self._constraint.short_description()


class WithDescription(Constraint):
    """Constraint that wraps another constraint and replaces its description

    Whenever a constraint's self-description does not fit an application
    context, it can be wrapped with this class. The given synopsis and
    description of valid inputs replaces those of the wrapped constraint.
    """
    def __init__(self,
                 constraint: Constraint,
                 *,
                 input_synopsis: str | None = None,
                 input_description: str | None = None,
                 error_message: str | None = None,
                 input_synopsis_for_ds: str | None = None,
                 input_description_for_ds: str | None = None,
                 error_message_for_ds: str | None = None,
    ):
        """
        Parameters
        ----------
        constraint: Constraint
          Any ``Constraint`` subclass instance that will be used to validate
          values.
        input_synopsis: optional
          If given, text to be returned as the constraint's ``input_synopsis``.
          Otherwise the wrapped constraint's ``input_synopsis`` is returned.
        input_description: optional
          If given, text to be returned as the constraint's
          ``input_description``. Otherwise the wrapped constraint's
          ``input_description`` is returned.
        error_message: optional
          If given, replaces the error message of a ``ConstraintError``
          raised by the wrapped ``Constraint``. Only the message
          (template) is replaced, not the error context dictionary.
        input_synopsis_for_ds: optional
          If either this, or ``input_description_for_ds``, or
          ``error_message_for_ds`` are given, the result of tailoring a
          constraint for a particular dataset (``for_dataset()``) will
          also be wrapped with this custom synopsis.
        input_description_for_ds: optional
          If either this, or ``input_synopsis_for_ds``, or
          ``error_message_for_ds`` are given, the result of tailoring a
          constraint for a particular dataset (``for_dataset()``) will
          also be wrapped with this custom description.
        error_message: optional
          If either this, or ``input_synopsis_for_ds``, or
          ``input_description_for_ds`` are given, the result of tailoring a
          constraint for a particular dataset (``for_dataset()``) will
          also be wrapped with this custom error message (template).
        """
        super().__init__()
        self._constraint = constraint
        self._synopsis = input_synopsis
        self._description = input_description
        self._error_message = error_message
        self._synopsis_for_ds = input_synopsis_for_ds
        self._description_for_ds = input_description_for_ds
        self._error_message_for_ds = error_message_for_ds

    @property
    def constraint(self) -> Constraint:
        """Returns the wrapped constraint instance"""
        return self._constraint

    def __call__(self, value) -> Any:
        try:
            return self._constraint(value)
        except ConstraintError as e:
            # rewrap the error to get access to the top-level
            # self-description.
            msg, cnstr, value, ctx = e.args
            raise ConstraintError(
                self,
                value,
                self._error_message or msg,
                ctx,
            ) from e

    def __str__(self) -> str:
        return \
            f'<{self._constraint.__class__.__name__} with custom description>'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}' \
               f'({self._constraint!r}, ' \
               f'input_synopsis={self._synopsis!r}, ' \
               f'input_description={self._description!r}, ' \
               f'input_synopsis_for_ds={self._synopsis_for_ds!r}, ' \
               f'input_description_for_ds={self._description_for_ds!r}, ' \
               f'error_message={self._error_message!r}, ' \
               f'error_message_for_ds={self._error_message_for_ds!r})'

    def for_dataset(self, dataset: DatasetParameter) -> Constraint:
        """Wrap the wrapped constraint again after tailoring it for the dataset
        """
        if any(x is not None for x in (
                self._synopsis_for_ds,
                self._description_for_ds,
                self._error_message_for_ds)):
            # we also want to wrap the tailored constraint
            return self.__class__(
                self._constraint.for_dataset(dataset),
                input_synopsis=self._synopsis_for_ds,
                input_description=self._description_for_ds,
                error_message=self._error_message_for_ds,
            )
        else:
            return self._constraint.for_dataset(dataset)

    @property
    def input_synopsis(self):
        return self._synopsis or self.constraint.input_synopsis

    @property
    def input_description(self):
        return self._description or self.constraint.input_description

    # legacy compatibility
    def long_description(self) -> str:
        return self.input_description

    def short_description(self) -> str:
        return self.input_synopsis



"""This file contains an updated :class:`CreateSiblingRia`-class.

It replaces the :class:`CreateSiblingRia`-class in
`datalad.distributed.create_sibling_ria`. The updated class uses `PurePosixPath`
for all RIA-store path calculations. It uses `url2transport_path` to convert
the abstract paths to concrete paths for the IO-abstraction.

The updated class also uses a canonified representation of path-anchors in
URL-paths. This allows to handle the differences in path-anchor encoding between
git-annex and RFC 8089-style file-URLs.
"""
from __future__ import annotations

import logging
from pathlib import PurePosixPath

from datalad.cmd import WitlessRunner as Runner
from datalad.core.distributed.clone import decode_source_spec
from datalad.customremotes.ria_utils import (
    create_ds_in_store,
    get_layout_locations,
    verify_ria_url,
)
from datalad.distributed.ora_remote import (
    LocalIO,
    RemoteCommandFailedError,
    RIARemoteError,
    SSHRemoteIO,
)
from datalad.distribution.dataset import (
    EnsureDataset,
    datasetmethod,
    require_dataset,
)
from datalad.distribution.utils import _yield_ds_w_matching_siblings
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results,
)
from datalad.interface.common_opts import (
    recursion_flag,
    recursion_limit,
)
from datalad.interface.results import get_status_dict
from datalad.log import log_progress
from datalad.support.annexrepo import AnnexRepo
from datalad.support.constraints import (
    EnsureBool,
    EnsureChoice,
    EnsureNone,
    EnsureStr,
)
from datalad.support.exceptions import CommandError
from datalad.support.gitrepo import GitRepo
from datalad.support.param import Parameter
from datalad.utils import quote_cmdlinearg

from datalad_next.consts import on_windows
from . import apply_patch
from .replace_ora_remote import (
    canonify_url,
    de_canonify_url,
)


lgr = logging.getLogger('datalad.distributed.create_sibling_ria')


# `CreateSiblingRia` taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732:
@build_doc
class CreateSiblingRia(Interface):
    """Creates a sibling to a dataset in a RIA store

    Communication with a dataset in a RIA store is implemented via two
    siblings. A regular Git remote (repository sibling) and a git-annex
    special remote for data transfer (storage sibling) -- with the former
    having a publication dependency on the latter. By default, the name of the
    storage sibling is derived from the repository sibling's name by appending
    "-storage".

    The store's base path is expected to not exist, be an empty directory,
    or a valid RIA store.

    Notes
    -----


    **RIA URL format**

    Interactions with new or existing RIA stores require RIA URLs to identify
    the store or specific datasets inside of it.

    The general structure of a RIA URL pointing to a store takes the form
    ``ria+[scheme]://<storelocation>`` (e.g.,
    ``ria+ssh://[user@]hostname:/absolute/path/to/ria-store``, or
    ``ria+file:///absolute/path/to/ria-store``)

    The general structure of a RIA URL pointing to a dataset in a store (for
    example for cloning) takes a similar form, but appends either the datasets
    UUID or a "~" symbol followed by the dataset's alias name:
    ``ria+[scheme]://<storelocation>#<dataset-UUID>`` or
    ``ria+[scheme]://<storelocation>#~<aliasname>``.
    In addition, specific version identifiers can be appended to the URL with an
    additional "@" symbol:
    ``ria+[scheme]://<storelocation>#<dataset-UUID>@<dataset-version>``,
    where ``dataset-version`` refers to a branch or tag.


    **RIA store layout**

    A RIA store is a directory tree with a dedicated subdirectory for each
    dataset in the store. The subdirectory name is constructed from the
    DataLad dataset ID, e.g. ``124/68afe-59ec-11ea-93d7-f0d5bf7b5561``, where
    the first three characters of the ID are used for an intermediate
    subdirectory in order to mitigate files system limitations for stores
    containing a large number of datasets.

    By default, a dataset in a RIA store consists of two components:
    A Git repository (for all dataset contents stored in Git) and a
    storage sibling (for dataset content stored in git-annex).

    It is possible to selectively disable either component using
    ``storage-sibling 'off'`` or ``storage-sibling 'only'``, respectively.
    If neither component is disabled, a dataset's subdirectory layout in a RIA
    store contains a standard bare Git repository and an ``annex/`` subdirectory
    inside of it.
    The latter holds a Git-annex object store and comprises the storage sibling.
    Disabling the standard git-remote (``storage-sibling='only'``) will result
    in not having the bare git repository, disabling the storage sibling
    (``storage-sibling='off'``) will result in not having the ``annex/``
    subdirectory.

    Optionally, there can be a further subdirectory ``archives`` with
    (compressed) 7z archives of annex objects. The storage remote is able to
    pull annex objects from these archives, if it cannot find in the regular
    annex object store. This feature can be useful for storing large
    collections of rarely changing data on systems that limit the number of
    files that can be stored.

    Each dataset directory also contains a ``ria-layout-version`` file that
    identifies the data organization (as, for example, described above).

    Lastly, there is a global ``ria-layout-version`` file at the store's
    base path that identifies where dataset subdirectories themselves are
    located. At present, this file must contain a single line stating the
    version (currently "1"). This line MUST end with a newline character.

    It is possible to define an alias for an individual dataset in a store by
    placing a symlink to the dataset location into an ``alias/`` directory
    in the root of the store. This enables dataset access via URLs of format:
    ``ria+<protocol>://<storelocation>#~<aliasname>``.

    Compared to standard git-annex object stores, the ``annex/`` subdirectories
    used as storage siblings follow a different layout naming scheme
    ('dirhashmixed' instead of 'dirhashlower').
    This is mostly noted as a technical detail, but also serves to remind
    git-annex powerusers to refrain from running git-annex commands
    directly in-store as it can cause severe damage due to the layout
    difference. Interactions should be handled via the ORA special remote
    instead.


    **Error logging**

    To enable error logging at the remote end, append a pipe symbol and an "l"
    to the version number in ria-layout-version (like so: ``1|l\\n``).

    Error logging will create files in an "error_log" directory whenever the
    git-annex special remote (storage sibling) raises an exception, storing the
    Python traceback of it. The logfiles are named according to the scheme
    ``<dataset id>.<annex uuid of the remote>.log`` showing "who" ran into this
    issue with which dataset. Because logging can potentially leak personal
    data (like local file paths for example), it can be disabled client-side
    by setting the configuration variable
    ``annex.ora-remote.<storage-sibling-name>.ignore-remote-config``.
    """

    # TODO: description?
    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""specify the dataset to process.  If
            no dataset is given, an attempt is made to identify the dataset
            based on the current working directory""",
            constraints=EnsureDataset() | EnsureNone()),
        url=Parameter(
            args=("url",),
            metavar="ria+<ssh|file|http(s)>://<host>[/path]",
            doc="""URL identifying the target RIA store and access protocol. If
            ``push_url||--push-url`` is given in addition, this is
            used for read access only. Otherwise it will be used for write
            access too and to create the repository sibling in the RIA store.
            Note, that HTTP(S) currently is valid for consumption only thus
            requiring to provide ``push_url||--push-url``.
            """,
            constraints=EnsureStr() | EnsureNone()),
        push_url=Parameter(
            args=("--push-url",),
            metavar="ria+<ssh|file>://<host>[/path]",
            doc="""URL identifying the target RIA store and access protocol for
            write access to the storage sibling. If given this will also be used
            for creation of the repository sibling in the RIA store.""",
            constraints=EnsureStr() | EnsureNone()),
        name=Parameter(
            args=('-s', '--name',),
            metavar='NAME',
            doc="""Name of the sibling.
            With `recursive`, the same name will be used to label all
            the subdatasets' siblings.""",
            constraints=EnsureStr() | EnsureNone(),
            required=True),
        storage_name=Parameter(
            args=("--storage-name",),
            metavar="NAME",
            doc="""Name of the storage sibling (git-annex special remote).
            Must not be identical to the sibling name. If not specified,
            defaults to the sibling name plus '-storage' suffix. If only
            a storage sibling is created, this setting is ignored, and
            the primary sibling name is used.""",
            constraints=EnsureStr() | EnsureNone()),
        alias=Parameter(
            args=('--alias',),
            metavar='ALIAS',
            doc="""Alias for the dataset in the RIA store.
            Add the necessary symlink so that this dataset can be cloned from the RIA
            store using the given ALIAS instead of its ID.
            With `recursive=True`, only the top dataset will be aliased.""",
            constraints=EnsureStr() | EnsureNone()),
        post_update_hook=Parameter(
            args=("--post-update-hook",),
            doc="""Enable Git's default post-update-hook for the created
            sibling. This is useful when the sibling is made accessible via a
            "dumb server" that requires running 'git update-server-info'
            to let Git interact properly with it.""",
            action="store_true"),
        shared=Parameter(
            args=("--shared",),
            metavar='{false|true|umask|group|all|world|everybody|0xxx}',
            doc="""If given, configures the permissions in the
            RIA store for multi-users access.
            Possible values for this option are identical to those of
            `git init --shared` and are described in its documentation.""",
            constraints=EnsureStr() | EnsureBool() | EnsureNone()),
        group=Parameter(
            args=("--group",),
            metavar="GROUP",
            doc="""Filesystem group for the repository. Specifying the group is
            crucial when [CMD: --shared=group CMD][PY: shared="group" PY]""",
            constraints=EnsureStr() | EnsureNone()),
        storage_sibling=Parameter(
            args=("--storage-sibling",),
            dest='storage_sibling',
            metavar='MODE',
            constraints=EnsureChoice('only') | EnsureBool() | EnsureNone(),
            doc="""By default, an ORA storage sibling and a Git repository
            sibling are created ([CMD: on CMD][PY: True|'on' PY]).
            Alternatively, creation of the storage sibling can be disabled
            ([CMD: off CMD][PY: False|'off' PY]), or a storage sibling
            created only and no Git sibling
            ([CMD: only CMD][PY: 'only' PY]). In the latter mode, no Git
            installation is required on the target host."""),
        existing=Parameter(
            args=("--existing",),
            constraints=EnsureChoice('skip', 'error', 'reconfigure'),
            metavar='MODE',
            doc="""Action to perform, if a (storage) sibling is already
            configured under the given name and/or a target already exists.
            In this case, a dataset can be skipped ('skip'), an existing target
            repository be forcefully re-initialized, and the sibling
            (re-)configured ('reconfigure'), or the command be instructed to
            fail ('error').""", ),
        new_store_ok=Parameter(
            args=("--new-store-ok",),
            action='store_true',
            doc="""When set, a new store will be created, if necessary. Otherwise, a sibling
            will only be created if the url points to an existing RIA store.""",
        ),
        recursive=recursion_flag,
        recursion_limit=recursion_limit,
        trust_level=Parameter(
            args=("--trust-level",),
            metavar="TRUST-LEVEL",
            constraints=EnsureChoice('trust', 'semitrust', 'untrust', None),
            doc="""specify a trust level for the storage sibling. If not
            specified, the default git-annex trust level is used. 'trust'
            should be used with care (see the git-annex-trust man page).""",),
        disable_storage__=Parameter(
            args=("--no-storage-sibling",),
            dest='disable_storage__',
            doc="""This option is deprecated. Use '--storage-sibling off'
            instead.""",
            action="store_false"),
    )

    @staticmethod
    @datasetmethod(name='create_sibling_ria')
    @eval_results
    def __call__(url,
                 name,
                 *,  # note that `name` is required but not posarg in CLI
                 dataset=None,
                 storage_name=None,
                 alias=None,
                 post_update_hook=False,
                 shared=None,
                 group=None,
                 storage_sibling=True,
                 existing='error',
                 new_store_ok=False,
                 trust_level=None,
                 recursive=False,
                 recursion_limit=None,
                 disable_storage__=None,
                 push_url=None
                 ):
        if disable_storage__ is not None:
            import warnings
            warnings.warn("datalad-create-sibling-ria --no-storage-sibling "
                          "is deprecated, use --storage-sibling off instead.",
                          DeprecationWarning)
            # recode to new setup
            disable_storage__ = None
            storage_sibling = False

        if storage_sibling == 'only' and storage_name:
            lgr.warning(
                "Sibling name will be used for storage sibling in "
                "storage-sibling-only mode, but a storage sibling name "
                "was provided"
            )

        ds = require_dataset(
            dataset, check_installed=True, purpose='create RIA sibling(s)')
        res_kwargs = dict(
            ds=ds,
            action="create-sibling-ria",
            logger=lgr,
        )

        # parse target URL
        # Note: URL parsing is done twice ATM (for top-level ds). This can't be
        # reduced to single instance, since rewriting url based on config could
        # be different for subdatasets.
        # PATCH: use canonified representation of `url` and `push_url`
        url = canonify_url(url)
        push_url = canonify_url(push_url)

        try:
            ssh_host, url_base_path_str, rewritten_url = \
                verify_ria_url(push_url if push_url else url, ds.config)
        except ValueError as e:
            yield get_status_dict(
                status='error',
                message=str(e),
                **res_kwargs
            )
            return

        # PATCH: use `PurePosixPath` to represent the base path of the store
        url_base_path = PurePosixPath(url_base_path_str)

        if ds.repo.get_hexsha() is None or ds.id is None:
            raise RuntimeError(
                "Repository at {} is not a DataLad dataset, "
                "run 'datalad create [--force]' first.".format(ds.path))

        if not storage_sibling and storage_name:
            lgr.warning(
                "Storage sibling setup disabled, but a storage sibling name "
                "was provided"
            )

        if storage_sibling and not storage_name:
            storage_name = "{}-storage".format(name)

        if storage_sibling and name == storage_name:
            # leads to unresolvable, circular dependency with publish-depends
            raise ValueError("sibling names must not be equal")

        if not isinstance(url, str):
            raise TypeError("url is not a string, but %s" % type(url))

        # Query existing siblings upfront in order to fail early on
        # existing=='error', since misconfiguration (particularly of special
        # remotes) only to fail in a subdataset later on with that config, can
        # be quite painful.
        # TODO: messages - this is "create-sibling". Don't confuse existence of
        #       local remotes with existence of the actual remote sibling
        #       in wording
        if existing == 'error':
            failed = False
            for dpath, sname in _yield_ds_w_matching_siblings(
                    ds,
                    (name, storage_name),
                    recursive=recursive,
                    recursion_limit=recursion_limit):
                res = get_status_dict(
                    status='error',
                    message=(
                        "a sibling %r is already configured in dataset %r",
                        sname, dpath),
                    type='sibling',
                    name=sname,
                    **res_kwargs,
                )
                failed = True
                yield res
            if failed:
                return
        # TODO: - URL parsing + store creation needs to be RF'ed based on
        #         command abstractions
        #       - more generally consider store creation a dedicated command or
        #         option

        io = SSHRemoteIO(ssh_host) if ssh_host else LocalIO()
        try:
            # determine the existence of a store by trying to read its layout.
            # Because this raises a FileNotFound error if non-existent, we need
            # to catch it
            # PATCH: convert ria-layout-version location to a concreate path
            # before giving it to an IO-abstraction.
            io.read_file(io.url2transport_path(url_base_path / 'ria-layout-version'))
        except (FileNotFoundError, RIARemoteError, RemoteCommandFailedError) as e:
            if not new_store_ok:
                # we're instructed to only act in case of an existing RIA store
                res = get_status_dict(
                    status='error',
                    message="No store found at '{}'. Forgot "
                            "--new-store-ok ?".format(url_base_path),
                    **res_kwargs)
                yield res
                return

        log_progress(
            lgr.info, 'create-sibling-ria',
            'Creating a new RIA store at %s', url_base_path,
        )
        create_store(io, url_base_path, '1')

        yield from _create_sibling_ria(
            ds,
            url,
            push_url,
            name,
            storage_sibling,
            storage_name,
            alias,
            existing,
            shared,
            group,
            post_update_hook,
            trust_level,
            res_kwargs)

        if recursive:
            # Note: subdatasets can be treated independently, so go full
            # recursion when querying for them and _no_recursion with the
            # actual call. Theoretically this can be parallelized.

            for subds in ds.subdatasets(state='present',
                                        recursive=True,
                                        recursion_limit=recursion_limit,
                                        return_type='generator',
                                        result_renderer='disabled',
                                        result_xfm='datasets'):
                yield from _create_sibling_ria(
                    subds,
                    url,
                    push_url,
                    name,
                    storage_sibling,
                    storage_name,
                    None,  # subdatasets can't have the same alias as the parent
                    existing,
                    shared,
                    group,
                    post_update_hook,
                    trust_level,
                    res_kwargs)


def _create_sibling_ria(
        ds,
        url,
        push_url,
        name,
        storage_sibling,
        storage_name,
        alias,
        existing,
        shared,
        group,
        post_update_hook,
        trust_level,
        res_kwargs):
    # be safe across datasets
    res_kwargs = res_kwargs.copy()
    # update dataset
    res_kwargs['ds'] = ds

    if not isinstance(ds.repo, AnnexRepo):
        # No point in dealing with a special remote when there's no annex.
        # Note, that in recursive invocations this might only apply to some of
        # the datasets. Therefore dealing with it here rather than one level up.
        lgr.debug("No annex at %s. Ignoring special remote options.", ds.path)
        storage_sibling = False
        storage_name = None

    # parse target URL
    try:
        ssh_host, url_base_path_str, rewritten_url = \
            verify_ria_url(push_url if push_url else url, ds.config)
    except ValueError as e:
        yield get_status_dict(
            status='error',
            message=str(e),
            **res_kwargs
        )
        return

    # PATCH: use `PurePosixPath` to represent the base path of the RIA-store
    url_base_path = PurePosixPath(url_base_path_str)

    git_url = decode_source_spec(
        # append dataset id to url and use magic from clone-helper:
        url + '#{}'.format(ds.id),
        cfg=ds.config
    )['giturl']
    git_push_url = decode_source_spec(
        push_url + '#{}'.format(ds.id),
        cfg=ds.config
    )['giturl'] if push_url else None

    # determine layout locations; go for a v1 store-level layout
    repo_path, _, _ = get_layout_locations(1, url_base_path, ds.id)

    ds_siblings = [
        r['name'] for r in ds.siblings(
            result_renderer='disabled',
            return_type='generator')
    ]
    # Figure whether we are supposed to skip this very dataset
    if existing == 'skip' and (
            name in ds_siblings or (
                storage_name and storage_name in ds_siblings)):
        yield get_status_dict(
            status='notneeded',
            message="Skipped on existing sibling",
            **res_kwargs
        )
        # if we skip here, nothing else can change that decision further
        # down
        return

    # figure whether we need to skip or error due an existing target repo before
    # we try to init a special remote.
    if ssh_host:
        from datalad import ssh_manager
        ssh = ssh_manager.get_connection(
            ssh_host,
            use_remote_annex_bundle=False)
        ssh.open()

    exists = False
    if existing in ['skip', 'error', 'reconfigure']:
        config_path = repo_path / 'config'
        # No .git -- if it's an existing repo in a RIA store it should be a
        # bare repo.
        # Theoretically we could have additional checks for whether we have
        # an empty repo dir or a non-bare repo or whatever else.
        if ssh_host:
            try:
                ssh('[ -e {p} ]'.format(p=quote_cmdlinearg(str(config_path))))
                exists = True
            except CommandError:
                exists = False
        else:
            # PATCH: use concrete path for `config_path` to check for existence
            exists = LocalIO().url2transport_path(config_path).exists()

        if exists:
            if existing == 'skip':
                # 1. not rendered by default
                # 2. message doesn't show up in ultimate result
                #    record as shown by -f json_pp
                yield get_status_dict(
                    status='notneeded',
                    message="Skipped on existing remote "
                            "directory {}".format(repo_path),
                    **res_kwargs
                )
                return
            elif existing == 'error':
                yield get_status_dict(
                    status='error',
                    message="remote directory {} already "
                            "exists.".format(repo_path),
                    **res_kwargs
                )
                return
            else:
                # reconfigure will be handled later in the code
                pass

    if storage_sibling == 'only':
        lgr.info("create storage sibling '%s' ...", name)
    else:
        lgr.info("create sibling%s '%s'%s ...",
            's' if storage_name else '',
            name,
            " and '{}'".format(storage_name) if storage_name else '',
        )
    create_ds_in_store(SSHRemoteIO(ssh_host) if ssh_host else LocalIO(),
                       # PATCH: use abstract `url_base_path`
                       url_base_path, ds.id, '2', '1', alias,
                       init_obj_tree=storage_sibling is not False)
    if storage_sibling:
        # we are using the main `name`, if the only thing we are creating
        # is the storage sibling
        srname = name if storage_sibling == 'only' else storage_name

        lgr.debug('init special remote %s', srname)
        special_remote_options = [
            'type=external',
            'externaltype=ora',
            'encryption=none',
            'autoenable=true',
            # PATCH: de-canonify `url`, because git-annex expects the path
            # anchor in the netloc position.
            f'url={de_canonify_url(url)}']
        if push_url:
            # PATCH: de-canonify `push_url`, because git-annex expects the path
            # anchor in the netloc position.
            special_remote_options.append(f'push-url={de_canonify_url(push_url)}')
        try:
            ds.repo.init_remote(
                srname,
                options=special_remote_options)
        except CommandError as e:
            if existing == 'reconfigure' \
                    and 'There is already a special remote' \
                    in e.stderr:
                # run enableremote instead
                lgr.debug(
                    "special remote '%s' already exists. "
                    "Run enableremote instead.",
                    srname)
                # TODO: Use AnnexRepo.enable_remote (which needs to get
                #       `options` first)
                ds.repo.call_annex([
                    'enableremote',
                    srname] + special_remote_options)
            else:
                yield get_status_dict(
                    status='error',
                    message="initremote failed.\nstdout: %s\nstderr: %s"
                    % (e.stdout, e.stderr),
                    **res_kwargs
                )
                return

        if trust_level:
            trust_cmd = [trust_level]
            if trust_level == 'trust':
                # Following git-annex 8.20201129-73-g6a0030a11, using `git
                # annex trust` requires --force.
                trust_cmd.append('--force')
            ds.repo.call_annex(trust_cmd + [srname])
        # get uuid for use in bare repo's config
        uuid = ds.config.get("remote.{}.annex-uuid".format(srname))

    if storage_sibling == 'only':
        # we can stop here, the rest of the function is about setting up
        # the git remote part of the sibling
        yield get_status_dict(
            status='ok',
            **res_kwargs,
        )
        return

    # 2. create a bare repository in-store:

    lgr.debug("init bare repository")
    # TODO: we should prob. check whether it's there already. How?
    # Note: like the special remote itself, we assume local FS if no
    # SSH host is specified
    disabled_hook = repo_path / 'hooks' / 'post-update.sample'
    enabled_hook = repo_path / 'hooks' / 'post-update'

    if group:
        chgrp_cmd = "chgrp -R {} {}".format(
            quote_cmdlinearg(str(group)),
            quote_cmdlinearg(str(repo_path)))

    if ssh_host:
        ssh('cd {rootdir} && git init --bare{shared}'.format(
            rootdir=quote_cmdlinearg(str(repo_path)),
            shared=" --shared='{}'".format(
                quote_cmdlinearg(shared)) if shared else ''
        ))

        if storage_sibling:
            # write special remote's uuid into git-config, so clone can
            # which one it is supposed to be and enable it even with
            # fallback URL
            ssh("cd {rootdir} && git config datalad.ora-remote.uuid {uuid}"
                "".format(rootdir=quote_cmdlinearg(str(repo_path)),
                          uuid=uuid))

        if post_update_hook:
            ssh('mv {} {}'.format(quote_cmdlinearg(str(disabled_hook)),
                                  quote_cmdlinearg(str(enabled_hook))))

        if group:
            # Either repository existed before or a new directory was
            # created for it, set its group to a desired one if was
            # provided with the same chgrp
            ssh(chgrp_cmd)

        # finally update server
        if post_update_hook:
            # Conditional on post_update_hook, since one w/o the other doesn't
            # seem to make much sense.
            ssh('cd {rootdir} && git update-server-info'.format(
                rootdir=quote_cmdlinearg(str(repo_path))
            ))
    else:
        gr = GitRepo(
            # PATCH: convert `repo_path` to a concrete path before stringifying
            # it and giving it to `GitRepo`.
            str(LocalIO().url2transport_path(repo_path)),
            create=True,
            bare=True,
            shared=shared if shared else None
        )
        if exists and existing == 'reconfigure':
            # if the repo exists at the given path, the GitRepo would not
            # (re)-run git init, and just return an instance of GitRepo;
            # skip & error have been handled at this point
            gr.init(
                sanity_checks=False,
                init_options=["--bare"] + ([f"--shared={shared}"] if shared else []),
            )
        if storage_sibling:
            # write special remote's uuid into git-config, so clone can
            # which one it is supposed to be and enable it even with
            # fallback URL
            gr.config.add("datalad.ora-remote.uuid", uuid, scope='local')

        if post_update_hook:
            # PATCH: convert `disabled_hook` to a concrete path before renaming
            LocalIO().url2transport_path(disabled_hook).rename(enabled_hook)
        if group:
            if on_windows:
                # PATCH: skip group-handling on Windows because there is no
                # `chgrp`.
                lgr.warning(
                    "Group '%s' was provided, but chgrp is not available on "
                    "Windows. Skipping.", group
                )
            else:
                # No CWD needed here, since `chgrp` is expected to be found via PATH
                # and the path it's operating on is absolute (repo_path). No
                # repository operation involved.
                Runner().run(chgrp_cmd)
        # finally update server
        if post_update_hook:
            # Conditional on post_update_hook, since one w/o the other doesn't
            # seem to make much sense.
            gr.call_git(["update-server-info"])

    # add a git remote to the bare repository
    # Note: needs annex-ignore! Otherwise we might push into dirhash
    # lower annex/object tree instead of mixed, since it's a bare
    # repo. This in turn would be an issue, if we want to pack the
    # entire thing into an archive. Special remote will then not be
    # able to access content in the "wrong" place within the archive
    lgr.debug("set up git remote")
    if name in ds_siblings:
        # otherwise we should have skipped or failed before
        assert existing == 'reconfigure'
    ds.config.set(
        "remote.{}.annex-ignore".format(name),
        value="true",
        scope="local")
    yield from ds.siblings(
        'configure',
        name=name,
        # PATCH: convert `repo_path` as a concrete path before stringifying it
        url=str(LocalIO().url2transport_path(repo_path))
        if url.startswith("ria+file")
        else git_url,
        pushurl=git_push_url,
        recursive=False,
        # Note, that this should be None if storage_sibling was not set
        publish_depends=storage_name,
        result_renderer='disabled',
        return_type='generator',
        # Note, that otherwise a subsequent publish will report
        # "notneeded".
        fetch=True
    )

    yield get_status_dict(
        status='ok',
        **res_kwargs,
    )


# Replace the complete `CreateSiblingRia`-class
apply_patch(
    'datalad.distributed.create_sibling_ria',
    None,
    'CreateSiblingRia',
    CreateSiblingRia,
)


# Replace `create_sibling_ria`-method in `Dataset`. This is necessary on
# Windows. I don't know why.
def _adaptor_create_sibling_ria(self, *args, **kwargs):
    return CreateSiblingRia.__call__(
        *args,
        **{
            **kwargs,
            'dataset': self,
        }
    )
apply_patch(
    'datalad.distribution.dataset',
    'Dataset',
    'create_sibling_ria',
    _adaptor_create_sibling_ria,
)


class UnknownLayoutVersion(Exception):
    pass


known_versions_objt = ['1', '2']
# Dataset tree versions we introduced so far. This is about the layout of
# datasets in a RIA store
known_versions_dst = ['1']


def _ensure_version(io, base_url, version):
    """Check a store or dataset version and make sure it is declared

    Parameters
    ----------
    io: SSHRemoteIO or LocalIO
    base_url: PurePosixPath
      root path of a store or dataset
    version: str
      target layout version of the store (dataset tree)
    """
    version_file = base_url / 'ria-layout-version'
    if io.exists(io.url2transport_path(version_file)):
        existing_version = io.read_file(
            io.url2transport_path(version_file)
        ).split('|')[0].strip()
        if existing_version != version.split('|')[0]:
            # We have an already existing location with a conflicting version on
            # record.
            # Note, that a config flag after pipe symbol is fine.
            raise ValueError("Conflicting version found at target: {}"
                             .format(existing_version))
        else:
            # already exists, recorded version fits - nothing to do
            return
    # Note, that the following does create the base-path dir as well, since
    # mkdir has parents=True:
    io.mkdir(io.url2transport_path(base_url))
    io.write_file(io.url2transport_path(version_file), version)


def create_store(io, base_url, version):
    """Helper to create a RIA store

    Note, that this is meant as an internal helper and part of intermediate
    RF'ing. Ultimately should lead to dedicated command or option for
    create-sibling-ria.

    Parameters
    ----------
    io: SSHRemoteIO or LocalIO
      Respective execution instance.
      Note: To be replaced by proper command abstraction
    base_url: PurePosixPath
      root url path of the store
    version: str
      layout version of the store (dataset tree)
    """
    assert isinstance(base_url, PurePosixPath)

    # At store level the only version we know as of now is 1.
    if version not in known_versions_dst:
        raise UnknownLayoutVersion("RIA store layout version unknown: {}."
                                   "Supported versions: {}"
                                   .format(version, known_versions_dst))
    _ensure_version(io, base_url, version)
    error_logs = base_url / 'error_logs'
    io.mkdir(io.url2transport_path(error_logs))



"""Auto-deploy credentials when enabling special remotes

This is the companion of the ``annexRepo__enable_remote`` patch, and simply
removes the webdav-specific credential handling in ``siblings()``.
It is no longer needed, because credential deployment moved to a lower
layer, covering more special remote types.

Manual credential entry on ``enableremote`` is not implemented here, but easily
possible following the patterns from `datalad-annex::` and
``create_sibling_webdav()``
"""
import logging

from datalad_next.datasets import LegacyAnnexRepo as AnnexRepo
from datalad.support.exceptions import (
    AccessDeniedError,
    AccessFailedError,
    CapturedException,
)
from . import apply_patch


# use same logger as -core
lgr = logging.getLogger('datalad.distribution.siblings')


# This function is taken from datalad-core@2ed709613ecde8218a215dcb7d74b4a352825685
# datalad/distribution/siblings.py
# Changes
# - removed credential lookup for webdav-remotes
# - exception logging via CapturedException
def _enable_remote(ds, repo, name, res_kwargs, **unused_kwargs):
    result_props = dict(
        action='enable-sibling',
        path=ds.path,
        type='sibling',
        name=name,
        **res_kwargs)

    if not isinstance(repo, AnnexRepo):
        yield dict(
            result_props,
            status='impossible',
            message='cannot enable sibling of non-annex dataset')
        return

    if name is None:
        yield dict(
            result_props,
            status='error',
            message='require `name` of sibling to enable')
        return

    # get info on special remote
    sp_remotes = {
        v['name']: dict(v, uuid=k)
        for k, v in repo.get_special_remotes().items()
    }
    remote_info = sp_remotes.get(name, None)

    if remote_info is None:
        yield dict(
            result_props,
            status='impossible',
            message=("cannot enable sibling '%s', not known", name))
        return

    try:
        repo.enable_remote(name)
        result_props['status'] = 'ok'
    except (AccessDeniedError, AccessFailedError) as e:
        CapturedException(e)
        result_props['status'] = 'error'
        # TODO should use proper way of injecting exceptions in result records
        result_props['message'] = str(e)

    yield result_props


apply_patch(
    'datalad.distribution.siblings', None, '_enable_remote', _enable_remote)



from __future__ import annotations

from pathlib import Path

from datalad_next.exceptions import CapturedException
from datalad_next.runners import (
    CommandError,
    call_git_lines,
)


def get_worktree_head(
    path: Path,
) -> tuple[str | None, str | None]:
    """Returns the symbolic name of the worktree `HEAD` at the given path

    Returns
    -------
    tuple
      The first item is the symbolic name of the worktree `HEAD`, or `None`
      if there is no commit.
      The second item is the symbolic name of the "corresponding branch" in
      an adjusted-mode git-annex repository, or `None`.
    """
    try:
        HEAD = call_git_lines(
            # we add the pathspec disambiguator to get cleaner error messages
            # (and we only report the first item below, to take it off again)
            ['rev-parse', '-q', '--symbolic-full-name', 'HEAD', '--'],
            cwd=path,
            # we are doing error message parsing below, fix the language
            # to avoid making it even more fragile
            force_c_locale=True,
        )[0]
    except (NotADirectoryError, FileNotFoundError) as e:
        raise ValueError('path not found') from e
    except CommandError as e:
        CapturedException(e)
        if 'fatal: not a git repository' in e.stderr:
            raise ValueError(f'no Git repository at {path!r}') from e
        elif 'fatal: bad revision' in e.stderr:
            return (None, None)
        else:
            # no idea reraise
            raise

    if HEAD.startswith('refs/heads/adjusted/'):
        # this is a git-annex adjusted branch. do the comparison against
        # its basis. it is not meaningful to track the managed branch in
        # a superdataset
        return (
            HEAD,
            # replace 'refs/heads' with 'refs/basis'
            f'refs/basis/{HEAD[11:]}',
        )
    else:
        return (HEAD, None)




from __future__ import annotations

from abc import ABCMeta
from datasalad.iterable_subprocess.iterable_subprocess import OutputFrom
from logging import getLogger

from ..response_generators import ShellCommandResponseGenerator


lgr = getLogger('datalad.ext.next.shell.operations')


class DownloadResponseGenerator(ShellCommandResponseGenerator, metaclass=ABCMeta):
    """Response generator interface for efficient download

    This response generator is used to implement download in a single command
    call (instead of using one command to determine the length of a file and
    a subsequent fixed-length command to download the file). It assumes that
    the shell sends ``<length>\\n``, the content of the
    file, and ``<return code>\\n``. The response generator delegates the
    creation of the appropriate final command list to its subclasses.
    """
    def __init__(self,
                 stdout: OutputFrom,
                 ) -> None:
        super().__init__(stdout, stdout.stderr_deque)
        self.length = 0
        self.read = 0
        self.state = 1
        self.returncode_chunk = b''

    def send(self, _) -> bytes:
        chunk = b''
        # Use a while loop to make arbitrary order of state checks possible.
        # This allows us to put the most active state at the top of the loop
        # and increase performance.
        while True:
            if self.state == 2:
                if not chunk:
                    chunk = next(self.stdout_gen)
                self.read += len(chunk)
                if self.read >= self.length:
                    self.state = 3
                    excess = self.read - self.length
                    if excess > 0:
                        chunk, self.returncode_chunk = chunk[:-excess], chunk[-excess:]
                    else:
                        self.returncode_chunk = b''
                    if chunk:
                        return chunk
                else:
                    return chunk

            if self.state == 1:
                self.length, chunk = self._get_number_and_newline(
                    b'',
                    self.stdout_gen,
                )
                # a negative length indicates an error during download length
                # determination or download length-communication.
                if self.length < 0:
                    self.state = 1
                    self.returncode = 23
                    raise StopIteration
                self.state = 2
                continue

            if self.state == 3:
                self.returncode, trailing = self._get_number_and_newline(
                    self.returncode_chunk,
                    self.stdout_gen,
                )
                if trailing:
                    lgr.warning(
                        'unexpected output after return code: %s',
                        repr(trailing))
                self.state = 4

            if self.state == 4:
                self.state = 1
                raise StopIteration

            raise RuntimeError(f'unknown state: {self.state}')



import contextlib
from os import environ

# legacy import
from datalad_next.patches import apply_patch


@contextlib.contextmanager
def patched_env(**env):
    """Context manager for patching the process environment

    Any number of kwargs can be given. Keys represent environment variable
    names, and values their values. A value of ``None`` indicates that
    the respective variable should be unset, i.e., removed from the
    environment.
    """
    preserve = {}
    for name, val in env.items():
        preserve[name] = environ.get(name, None)
        if val is None:
            del environ[name]
        else:
            environ[name] = str(val)
    try:
        yield
    finally:
        for name, val in preserve.items():
            if val is None:
                del environ[name]
            else:
                environ[name] = val



"""Handler for operations on various archive types

All handlers implement the API defined by :class:`ArchiveOperations`.

Available handlers:

.. currentmodule:: datalad_next.archive_operations
.. autosummary::
   :toctree: generated

   TarArchiveOperations
   ZipArchiveOperations
"""
from .tarfile import TarArchiveOperations
from .zipfile import ZipArchiveOperations


# TODO REMOVE EVERYTHING BELOW FOR V2.0
import logging
from abc import (
    ABC,
    abstractmethod,
)
from contextlib import contextmanager
from typing import (
    Any,
    Generator,
    IO,
)
# this API is not cooked enough yet to promote it for 3rd-part extensions
from .base import ArchiveOperations
import datalad
from datalad_next.config import ConfigManager
from datalad_next.iter_collections.utils import FileSystemItem


lgr = logging.getLogger('datalad.ext.next.archive_operations')



from ..annex import has_initialized_annex


def test_has_initialized_annex(existing_dataset):
    # for the root
    assert has_initialized_annex(existing_dataset.pathobj)
    # for a subdir
    assert has_initialized_annex(existing_dataset.pathobj / '.datalad')


def test_no_initialized_annex(existing_noannex_dataset, tmp_path):
    # for the root
    assert not has_initialized_annex(existing_noannex_dataset.pathobj)
    # for a subdir
    assert not has_initialized_annex(
        existing_noannex_dataset.pathobj / '.datalad')
    # for a random directory
    assert not has_initialized_annex(tmp_path)



"""Report on the content of directories

The main functionality is provided by the :func:`iter_dir()` function.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Generator

from datalad_next.exceptions import CapturedException

from .utils import (
    FileSystemItem, FileSystemItemType)


@dataclass  # sadly PY3.10+ only (kw_only=True)
class DirectoryItem(FileSystemItem):
    pass


def iter_dir(
    path: Path,
    *,
    fp: bool = False,
) -> Generator[DirectoryItem, None, None]:
    """Uses ``Path.iterdir()`` to iterate over a directory and reports content

    The iterator produces an :class:`DirectoryItem` instance with standard
    information on file system elements, such as ``size``, or ``mtime``.

    In addition to a plain ``Path.iterdir()`` the report includes a path-type
    label (distinguished are ``file``, ``directory``, ``symlink``).

    Parameters
    ----------
    path: Path
      Path of the directory to report content for (iterate over).
    fp: bool, optional
      If ``True``, each file-type item includes a file-like object
      to access the file's content. This file handle will be closed
      automatically when the next item is yielded.

    Yields
    ------
    :class:`DirectoryItem`
      The ``name`` attribute of an item is a ``Path`` instance, with the
      format matching the main ``path`` argument. When an absolute ``path``
      is given, item names are absolute paths too. When a relative path is
      given, it is relative to CWD, and items names are relative paths
      (relative to CWD) too.
    """
    for c in path.iterdir():
        # c could disappear while this is running. Example: temp files managed
        # by other processes.
        try:
            item = DirectoryItem.from_path(
                c,
                link_target=True,
            )
        except FileNotFoundError as e:
            CapturedException(e)
            continue
        if fp and item.type == FileSystemItemType.file:
            with c.open('rb') as fileobj:
                item.fp = fileobj
                yield item
        else:
            yield item



from pathlib import (
    Path,
    PurePath,
)
import pytest

from datalad_next.datasets import Dataset

from ..results import (
    CommandResult,
    CommandResultStatus,
)


def test_commandresult():
    # CommandResult is a plain data class, so there is no much to test
    # only the partial dict API that is implemented as a compatibility
    # shim for the datalad core result loop
    #
    # we need action, status, and path unconditionally
    with pytest.raises(TypeError):
        CommandResult()
    with pytest.raises(TypeError):
        CommandResult(action='some')
    with pytest.raises(TypeError):
        CommandResult(action='some', status='ok')
    # no something that works
    st = CommandResult(
        action='actionlabel',
        status=CommandResultStatus.ok,
        path=PurePath('mypath'),
        refds=Dataset('myds'),
    )
    # we can get a dict with stringified values (for some types)
    assert dict(st.items()) == {
        'action': 'actionlabel',
        'status': 'ok',
        'path': 'mypath',
        'message': None,
        'exception': None,
        'error_message': None,
        'type': None,
        'logger': None,
        'refds': str(Path.cwd() / 'myds'),
    }
    # 'in' works
    assert 'action' in st
    assert 'weird' not in st
    # getitem works, and gives strings
    assert st['path'] == 'mypath'
    # same for get
    assert st.get('path') == 'mypath'
    assert st.get('weird', 'normal') == 'normal'
    # 'pop' is emulated by setting to None
    assert st.pop('path') == 'mypath'
    assert st.path is None
    # popping something unknown doesn't blow
    assert st.pop('weird', 'default') == 'default'
    # and does not add cruft to the dataclass instance
    assert not hasattr(st, 'weird')




"""Report on the difference of two Git tree-ishes or tracked worktree content

The main functionality is provided by the :func:`iter_gitdiff()` function.
"""
from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
import logging
from pathlib import (
    Path,
    PurePosixPath,
)
from typing import (
    Iterator,
    List,
    Generator,
)
from datasalad.itertools import (
    decode_bytes,
    itemize,
)

from datalad_next.consts import PRE_INIT_COMMIT_SHA
from datasalad.gitpathspec import GitPathSpecs
from datalad_next.runners import (
    CommandError,
    iter_git_subproc,
)
from datalad_next.runners import (
    call_git,
    call_git_oneline,
)

from .gittree import (
    GitTreeItem,
    GitTreeItemType,
    _mode_type_map,
)

lgr = logging.getLogger('datalad.ext.next.iter_collections.gitdiff')


# TODO Could be `StrEnum`, came with PY3.11
class GitDiffStatus(Enum):
    """Enumeration of statuses for diff items
    """
    addition = 'addition'
    copy = 'copy'
    deletion = 'deletion'
    modification = 'modification'
    rename = 'rename'
    typechange = 'typechange'
    unmerged = 'unmerged'
    unknown = 'unknown'
    # this is a local addition and not defined by git
    # AKA "untracked"
    other = 'other'


_diffstatus_map = {
    'A': GitDiffStatus.addition,
    'C': GitDiffStatus.copy,
    'D': GitDiffStatus.deletion,
    'M': GitDiffStatus.modification,
    'R': GitDiffStatus.rename,
    'T': GitDiffStatus.typechange,
    'U': GitDiffStatus.unmerged,
    'X': GitDiffStatus.unknown,
    'O': GitDiffStatus.other,
}


# TODO Could be `StrEnum`, came with PY3.11
class GitContainerModificationType(Enum):
    new_commits = 'new commits'
    untracked_content = 'untracked content'
    modified_content = 'modified content'


@dataclass
class GitDiffItem(GitTreeItem):
    """``GitTreeItem`` with "previous" property values given a state comparison
    """
    prev_name: str | None = None
    prev_gitsha: str | None = None
    prev_gittype: GitTreeItemType | None = None

    status: GitDiffStatus | None = None
    percentage: int | None = None
    """This is the percentage of similarity for copy-status and
    rename-status diff items, and the percentage of dissimilarity
    for modifications."""
    modification_types: tuple[GitContainerModificationType, ...] | None = None
    """Qualifiers for modification types of container-type
    items (directories, submodules)."""

    def __post_init__(self) -> None:
        if self.status == GitDiffStatus.addition and self.gitsha is None:
            self.add_modification_type(GitContainerModificationType.modified_content)

    @cached_property
    def prev_path(self) -> PurePosixPath | None:
        """Returns the item ``prev_name`` as a ``PurePosixPath``
        instance"""
        if self.prev_name is None:
            return None
        return PurePosixPath(self.prev_name)

    def add_modification_type(
        self,
        value: GitContainerModificationType,
    ) -> None:
        if self.modification_types is None:
            self.modification_types = (value,)
        else:
            self.modification_types = (*self.modification_types, value)


def iter_gitdiff(
    path: Path,
    from_treeish: str | None,
    to_treeish: str | None,
    *,
    recursive: str = 'repository',
    find_renames: int | None = None,
    find_copies: int | None = None,
    yield_tree_items: str | None = None,
    eval_submodule_state: str = 'full',
    pathspecs: list[str] | GitPathSpecs | None = None,
) -> Generator[GitDiffItem, None, None]:
    """Report differences between Git tree-ishes or tracked worktree content

    This function is a wrapper around the Git command ``diff-tree`` and
    ``diff-index``. Therefore most semantics also apply here.

    The main difference with respect to the Git commands are: 1) uniform
    support for non-recursive, single tree reporting (no subtrees); and
    2) support for submodule recursion.

    Notes on 'no' recursion mode

    When comparing to the worktree, ``git diff-index`` always reports on
    subdirectories. For homogeneity with the report on a committed tree,
    a non-recursive mode emulation is implemented. It compresses all reports
    from a direct subdirectory into a single report on that subdirectory.
    The ``gitsha`` of that directory item will always be ``None``. Moreover,
    no type or typechange inspection, or further filesystem queries are
    performed. Therefore, ``prev_gittype`` will always be ``None``, and
    any change other than the addition of the directory will be labeled
    as a ``GitDiffStatus.modification``.

    Parameters
    ----------
    path: Path
      Path of a directory in a Git repository to report on. This directory
      need not be the root directory of the repository, but must be part of
      the repository. If the directory is not the root directory of a
      non-bare repository, the iterator is constrained to items underneath
      that directory.
    from_treeish: str or None
      Git "tree-ish" that defines the comparison reference. If ``None``,
      ``to_treeeish`` must not be ``None`` (see its documentation for
      details).
    to_treeish:
      Git "tree-ish" that defines the comparison target. If ``None``,
      ``from_treeish`` must not be ``None``, and that tree-ish will be
      compared against the worktree. (see its documentation for
      details). If ``from_treeish`` is ``None``, the given tree-ish is
      compared to its immediate parents (see ``git diff-tree`` documentation
      for details).
    recursive: {'repository', 'submodules', 'no'}, optional
      Behavior for recursion into subtrees. By default (``repository``),
      all trees within the repository underneath ``path``) are reported,
      but no tree within submodules. With ``submodules``, recursion includes
      any submodule that is present. If ``no``, only direct children
      are reported on.
    find_renames: int, optional
      If given, this defines the similarity threshold for detecting renames
      (see ``git diff-{index,tree} --find-renames``). By default, no rename
      detection is done and reported items never have the ``rename`` status.
      Instead, a renames would be reported as a deletion and an addition.
    find_copied: int, optional
      If given, this defines the similarity threshold for detecting copies
      (see ``git diff-{index,tree} --find-copies``). By default, no copy
      detection is done and reported items never have the ``copy`` status.
      Instead, a copy would be reported as addition.
      This option always implies the use of the ``--find-copies-harder``
      Git option that enables reporting of copy sources, even when they
      have not been modified in the same change. This is a very expensive
      operation for large projects, so use it with caution.
    yield_tree_items: {'submodules', 'directories', 'all', None}, optional
      Whether to yield an item on type of subtree that will also be recursed
      into. For example, a submodule item, when submodule recursion is
      enabled. When disabled, subtree items (directories, submodules)
      will still be reported whenever there is no recursion into them.
      For example, submodule items are reported when
      ``recursive='repository``, even when ``yield_tree_items=None``.
    eval_submodule_state: {"no", "commit", "full"}, optional
      If 'full' (default), the state of a submodule is evaluated by
      considering all modifications ('--ignore-submodules=none').
      If 'commit', the modification check is restricted to comparing the
      submodule's "HEAD" commit to the one recorded in the superdataset
      ('--ignore-submodules=dirty'). If 'no', the state of the subdataset is
      not evaluated ('--ignore-submodules=all').
    pathspecs: optional
      Patterns used to limit results to particular paths. Any pathspecs
      supported by Git can be used and are passed to the underlying ``git
      ls-files`` queries. Pathspecs are also supported for recursive reporting
      on submodules. In such a case, the results match those of individual
      queries with analog pathspecs on the respective submodules (Git itself
      does not support pathspecs for submodule-recursive operations).  For
      example, a ``submodule`` recursion with a pathspec ``*.jpg`` will yield
      reports on all JPG files in all submodules, even though a submodule path
      itself does not match ``*.jpg``.  On the other hand, a pathspec
      ``submoddir/*.jpg`` will only report on JPG files in the submodule at
      ``submoddir/``, but on all JPG files in that submodule.
      As of version 1.5, the pathspec support for submodule recursion is
      preliminary and results should be carefully investigated.

    Yields
    ------
    :class:`GitDiffItem`
      The ``name`` and ``prev_name`` attributes of an item are a ``str`` with
      the corresponding (relative) path, as reported by Git
      (in POSIX conventions).
    """
    # we force-convert to Path to give us the piece of mind we want.
    # The docs already ask for that, but it is easy to
    # forget/ignore and leads to non-obvious errors. Running this once is
    # a cheap safety net
    path = Path(path)
    _pathspecs = GitPathSpecs(pathspecs)
    processed_submodules: set[str] = set()

    for item in _iter_gitdiff(
        path=path,
        from_treeish=from_treeish,
        to_treeish=to_treeish,
        recursive=recursive,
        find_renames=find_renames,
        find_copies=find_copies,
        yield_tree_items=yield_tree_items,
        eval_submodule_state=eval_submodule_state,
        pathspecs=_pathspecs,
    ):
        # exclude non-submodules, or a submodule that was found at
        # the root path -- which would indicate that the submodule
        # itself it not around, only its record in the parent
        if recursive == 'submodules' \
                and item.gittype == GitTreeItemType.submodule \
                and not item.name == '.':
            # mark as processed immediately, independent of whether anything
            # need to be reported
            processed_submodules.add(item.name)
        yield item

    # we may need to loop over the (remaining) submodules, because
    # with pathspecs there is a chance that a given pathspec set did not
    # match a submodule (directly) that could have content that matches a
    # pathspec
    if recursive == 'submodules' and _pathspecs:
        for item in _iter_gitdiff(
            path=path,
            from_treeish=from_treeish,
            to_treeish=to_treeish,
            # no need to double-recurse, we just need to discover all
            # submodules in the diff unconstrained by pathspecs
            recursive='repository',
            find_renames=None,
            find_copies=None,
            yield_tree_items=None,
            # we need to look at the recorded commit to get submodule
            # reports at all
            eval_submodule_state='commit',
            pathspecs=GitPathSpecs(None),
        ):
            if item.gittype != GitTreeItemType.submodule \
                    or item.name in processed_submodules:
                # not a submodule or already reported on
                continue

            yield from _yield_from_submodule(
                basepath=path,
                subm=item,
                to_treeish=to_treeish,
                recursive=recursive,
                yield_tree_items=yield_tree_items,
                find_renames=find_renames,
                find_copies=find_copies,
                eval_submodule_state=eval_submodule_state,
                pathspecs=_pathspecs,
            )


def _iter_gitdiff(
    path: Path,
    from_treeish: str | None,
    to_treeish: str | None,
    *,
    recursive: str,
    find_renames: int | None,
    find_copies: int | None,
    yield_tree_items: str | None,
    eval_submodule_state: str,
    pathspecs: GitPathSpecs,
) -> Generator[GitDiffItem, None, None]:
    cmd = _build_cmd(
        from_treeish=from_treeish,
        to_treeish=to_treeish,
        recursive=recursive,
        find_renames=find_renames,
        find_copies=find_copies,
        yield_tree_items=yield_tree_items,
        eval_submodule_state=eval_submodule_state,
        pathspecs=pathspecs,
    )

    if cmd[0] == 'diff-index':
        # when we compare to the index, we need a refresh run to not have
        # something like plain mtime changes trigger modification reports
        # https://github.com/datalad/datalad-next/issues/639
        call_git([
            'update-index',
            # must come first, we recurse ourselves
            '--ignore-submodules',
            # we want to continue the refresh when the index need updating
            '-q',
            '--refresh',
        ], cwd=path)

    # when do we need to condense subdir reports into a single dir-report
    reported_dirs: set[str] = set()
    _single_dir = (cmd[0] == 'diff-index') and recursive == 'no'
    # diff-tree reports the compared tree when no from is given, we need
    # to skip that output below
    skip_first = (cmd[0] == 'diff-tree') and from_treeish is None
    pending_props = None
    for line in _git_diff_something(path, cmd):
        if skip_first:
            skip_first = False
            continue
        if pending_props:
            pending_props.append(line)
            if pending_props[4][0] in ('C', 'R'):
                # for copies and renames we expect a second path
                continue
            yield from _yield_diff_item(
                cwd=path,
                single_dir=_single_dir,
                spec=pending_props,
                reported_dirs=reported_dirs,
                from_treeish=from_treeish,
                to_treeish=to_treeish,
                recursive=recursive,
                find_renames=find_renames,
                find_copies=find_copies,
                yield_tree_items=yield_tree_items,
                eval_submodule_state=eval_submodule_state,
                pathspecs=pathspecs,
            )
            pending_props = None
        elif line.startswith(':'):
            pending_props = line[1:].split(' ')
        else:  # pragma: no cover
            raise RuntimeError(
                'we should not get here, unexpected diff output')
    if pending_props:
        # flush
        yield from _yield_diff_item(
            cwd=path,
            single_dir=_single_dir,
            spec=pending_props,
            reported_dirs=reported_dirs,
            from_treeish=from_treeish,
            to_treeish=to_treeish,
            recursive=recursive,
            find_renames=find_renames,
            find_copies=find_copies,
            yield_tree_items=yield_tree_items,
            eval_submodule_state=eval_submodule_state,
            pathspecs=pathspecs,
        )


def _build_cmd(
    *,
    from_treeish: str | None,
    to_treeish: str | None,
    recursive: str,
    find_renames: int | None,
    find_copies: int | None,
    yield_tree_items: str | None,
    eval_submodule_state: str,
    pathspecs: GitPathSpecs,
) -> list[str]:
    # from   : to   : description
    # ---------------------------
    # HEAD   : None : compare to worktree, not with the index (diff-index)
    # HEAD~2 : HEAD : compare trees (diff-tree)
    # None   : HEAD~2 : compare tree with its parents (diff-tree)
    # None   : None : exception

    common_args: list[str] = [
        '--no-rename-empty',
        # ignore changes above CWD
        '--relative',
        '--raw',
        '-z',
    ]
    if find_renames is not None:
        common_args.append(f'--find-renames={find_renames}%')
    if find_copies is not None:
        common_args.append(f'--find-copies={find_copies}%')
        # if someone wants to look for copies, we actually look
        # for copies. This is expensive, but IMHO is the one
        # thing that makes this useful
        # TODO possibly we only want to enable this when
        # find_copies==100 (exact copies), based on the assumption
        # that this is cheaper than reading all file content.
        # but if that is actually true remains to be tested
        common_args.append(f'--find-copies-harder')

    if eval_submodule_state == 'no':
        common_args.append('--ignore-submodules=all')
    elif eval_submodule_state == 'commit':
        common_args.append('--ignore-submodules=dirty')
    elif eval_submodule_state == 'full':
        common_args.append('--ignore-submodules=none')
    else:
        raise ValueError(
            f'unknown submodule evaluation mode {eval_submodule_state!r}')

    if from_treeish is None and to_treeish is None:
        raise ValueError(
            'either `from_treeish` or `to_treeish` must not be None')
    elif to_treeish is None:
        assert from_treeish is not None
        cmd = ['diff-index', *common_args, from_treeish]
    else:
        # diff NOT against the working tree
        cmd = ['diff-tree', *common_args]
        if recursive in ('repository', 'submodules'):
            cmd.append('-r')
            if yield_tree_items in ('all', 'directories'):
                cmd.append('-t')
        if from_treeish is None:
            cmd.append(to_treeish)
        else:
            # two tree-ishes given
            cmd.extend((from_treeish, to_treeish))

    # add disambiguation marker for pathspec.
    # even if we do not pass any, we get simpler error messages from Git
    cmd.append('--')

    cmd.extend(pathspecs.arglist())
    return cmd


def _get_diff_item(spec: list[str]) -> GitDiffItem:
    # this type annotation is a crutch, we'd really need
    # https://peps.python.org/pep-0692/ to do this properly
    props: dict[str, str | int | GitTreeItemType | GitDiffStatus | None] = {}
    props.update(
        (k, _mode_type_map.get(v, None))
        for k, v in (('prev_gittype', spec[0]),
                     ('gittype', spec[1]))
    )
    props.update(
        (k, None if v == (40 * '0') else v)
        for k, v in (('prev_gitsha', spec[2]),
                     ('gitsha', spec[3]))
    )
    status = spec[4]
    props['status'] = _diffstatus_map[status[0]]
    if len(status) > 1:
        props['percentage'] = int(status[1:])

    if status == 'A':
        # this is an addition, we want `name` in the right place
        props['name'] = spec[5]
    else:
        props['prev_name'] = spec[5]
        props['name'] = spec[6] if len(spec) > 6 else spec[5]

    # disable type checking here, we need
    # https://peps.python.org/pep-0692/ to do this properly
    return GitDiffItem(**props)  # type: ignore


def _yield_diff_item(
    *,
    cwd: Path,
    recursive: str,
    from_treeish: str | None,
    to_treeish: str | None,
    spec: list,
    single_dir: bool,
    reported_dirs: set,
    yield_tree_items: str | None,
    find_renames: int | None,
    find_copies: int | None,
    eval_submodule_state: str,
    pathspecs: GitPathSpecs,
) -> Generator[GitDiffItem, None, None]:
    item = _get_diff_item(spec)

    if single_dir:
        # handle the special case of reporting only on the 1st-level
        # containing directory of an item.
        name = item.name or item.prev_name
        # we cannot have items that have no name whatsoever
        assert name is not None
        # we decide on mangling the actual report to be on the containing
        # directory only, or to withhold it entirely
        dname_l = name.split('/', maxsplit=1)
        if len(dname_l) < 2:
            # nothing in a subdirectory
            yield item
            return
        dname = dname_l[0]
        if dname in reported_dirs:
            # nothing else todo, we already reported
            return

        reported_dirs.add(dname)
        yield _mangle_item_for_singledir(item, dname, from_treeish, cwd)
        return

    if item.gittype != GitTreeItemType.submodule:
        # any non-submodule item can be yielded now and we are done here
        yield item
        return

    # this is about a present submodule
    if item.status == GitDiffStatus.modification:
        if item.gitsha is None:
            # in 'git diff-index' speak the submodule is "out-of-sync" with
            # the index: this happens when there are new commits
            item.add_modification_type(
                GitContainerModificationType.new_commits)
        # TODO we cannot give details for other modification types.
        # depending on --ignore-submodules a range of situations
        # could be the case
        #else:
        #    # this modification means that "content" is modified
        #    item.add_modification_type(
        #        GitContainerModificationType.modified_content)

    if recursive != 'submodules':
        # no submodule recursion, we can yield this submodule item
        # directly
        yield item
        return

    if yield_tree_items in ('all', 'submodules'):
        # we are instructed to yield this submodule item, but if we are going
        # to recurse into it, continuing to use the item instance for
        # queries. Hence we yield a copy here to avoid data corruption
        yield deepcopy(item) if recursive == 'submodules' else item

    if recursive == 'submodules':
        yield from _yield_from_submodule(
            basepath=cwd,
            subm=item,
            to_treeish=to_treeish,
            recursive=recursive,
            yield_tree_items=yield_tree_items,
            find_renames=find_renames,
            find_copies=find_copies,
            eval_submodule_state=eval_submodule_state,
            pathspecs=pathspecs,
        )


def _yield_from_submodule(
    *,
    basepath: Path,
    subm: GitDiffItem,
    to_treeish: str | None,
    recursive: str,
    yield_tree_items: str | None,
    find_renames: int | None,
    find_copies: int | None,
    eval_submodule_state: str,
    pathspecs: GitPathSpecs,
) -> Generator[GitDiffItem, None, None]:
    # I believe we need no protection against absent submodules.
    # The only way they can appear here is a reported modification.
    # The only modification that is possible with an absent submodule
    # is a deletion. And that would cause the item.gittype to be None
    # -- a condition that is caught above
    subm_name = PurePosixPath(subm.name)
    subm_pathspecs = pathspecs
    if pathspecs:
        # recode pathspecs to match the submodule scope
        try:
            subm_pathspecs = pathspecs.for_subdir(subm_name)
        except ValueError:
            # not a single pathspec could be translated, there is
            # no chance for a match, we can stop here
            return
    for i in iter_gitdiff(
        basepath / subm_name,
        # we never want to pass None here
        # if `prev_gitsha` is None, it means that the
        # submodule record is new, and we want its full
        # content reported. Passing None, however,
        # would only report the change to the current
        # state.
        from_treeish=subm.prev_gitsha or PRE_INIT_COMMIT_SHA,
        # when comparing the parent to the worktree, we
        # also want to compare any children to the worktree
        to_treeish=None if to_treeish is None else subm.gitsha,
        # pass on the common args
        recursive=recursive,
        yield_tree_items=yield_tree_items,
        find_renames=find_renames,
        find_copies=find_copies,
        eval_submodule_state=eval_submodule_state,
        pathspecs=subm_pathspecs,
    ):
        # prepend any item name with the parent items
        # name
        for attr in ('name', 'prev_name'):
            val = getattr(i, attr)
            if val is not None:
                setattr(i, attr, f'{subm.name}/{val}')
        yield i


def _mangle_item_for_singledir(
    item: GitDiffItem,
    dname: str,
    from_treeish: str | None,
    cwd: Path,
) -> GitDiffItem:
    # at this point we have a change report on subdirectory content
    # we only get here when comparing `from_treeish` to the worktree.
    item.name = dname
    # non-committed change -> no SHA (this ignored the index,
    # like we do elsewhere too)
    item.gitsha = None
    item.gittype = GitTreeItemType.directory
    try:
        item.prev_gitsha = call_git_oneline(
            ['rev-parse', '-q', f'{from_treeish}:./{dname}'],
            cwd=cwd,
        )
        # if we get here, we know that the name was valid in
        # `from_treeish` too
        item.prev_name = dname
        # it would require more calls to figure out the mode and infer
        # a possible type change. For now, we do not go there
        item.prev_gittype = None
        item.status = GitDiffStatus.modification
    except CommandError:
        # the was nothing with this name in `from_treeish`, but now
        # it exists. We compare to the worktree, but not any untracked
        # content -- this means that we likely compare across multiple
        # states and the directory become tracked after `from_treeish`.
        # let's call it an addition
        item.prev_gitsha = None
        item.prev_gittype = None
        item.status = GitDiffStatus.addition

    return item


def _git_diff_something(path: Path, args: List[str]) -> Iterator[str]:
    with iter_git_subproc([*args], cwd=path) as r:
        yield from itemize(
            decode_bytes(r),
            sep='\0',
            keep_ends=False,
        )



# Helper module to develop git-annex backends
#
# https://git-annex.branchable.com/design/external_backend_protocol/
#
# Derived from AnnexRemote Copyright (C) 2017  Silvio Ankermann (GPL-3)
"""Interface and essential utilities to implement external git-annex backends
"""

import logging

from abc import (
    ABCMeta,
    abstractmethod,
)

import sys
import traceback


class Backend(metaclass=ABCMeta):
    """Metaclass for backends.

    It implements the communication with git-annex via the external backend
    protocol. More information on the protocol is available at
    https://git-annex.branchable.com/design/external_backend_protocol/

    External backends can be built by implementing the abstract methods defined
    in this class.

    Attributes
    ----------
    annex : Master
        The Master object to which this backend is linked. Master acts as an
        abstraction layer for git-annex.
    """

    def __init__(self, annex):
        self.annex = annex

    @abstractmethod
    def can_verify(self):
        """Returns whether the backend can verify the content of files match a
        key it generated. The verification does not need to be
        cryptographically secure, but should catch data corruption.

        Returns
        -------
        bool
        """

    @abstractmethod
    def is_stable(self):
        """Returns whether a key it has generated will always have the same
        content. The answer to this is almost always yes; URL keys are an
        example of a type of key that may have different content at different
        times.

        Returns
        -------
        bool
        """

    @abstractmethod
    def is_cryptographically_secure(self):
        """ Returns whether keys it generates are verified using a
        cryptographically secure hash.

        Note that sha1 is not a cryptographically secure hash any longer.
        A program can change its answer to this question as the state of the
        art advances, and should aim to stay ahead of the state of the art by
        a reasonable amount of time.

        Returns
        -------
        bool
        """

    @abstractmethod
    def gen_key(self, local_file):
        """Examine the content of `local_file` and from it generate a key.

        While it is doing this, it can send any number of PROGRESS messages
        indication the position in the file that it's gotten to.

        Parameters
        ----------
        local_file: str
            Path for which to generate a key.
            Note that in some cases, local_file may contain whitespace.

        Returns
        -------
        str
          The generated key.

        Raises
        ------
        BackendError
            If the file could not be received from the backend.
        """

    @abstractmethod
    def verify_content(self, key, content_file):
        """Examine a file and verify it has the content expected given a key

        While it is doing this, it can send any number of PROGRESS messages
        indicating the position in the file that it's gotten to.

        If `can_verify() == False`, git-annex not ask to do this.


        Returns
        -------
        bool
        """

    def error(self, error_msg):
        """Communicate a generic error.

        Can be sent at any time if things get too messed up to
        continue.  If the program receives an error() from git-annex, it can
        exit with its own error().  Eg.: self.annex.error("Error received.
        Exiting.") raise SystemExit

        Parameters
        ----------
        error_msg : str
            The error message received from git-annex
        """
        self.annex.error("Error received. Exiting.")
        raise SystemExit


# Exceptions
class AnnexError(Exception):
    """
    Common base class for all annexbackend exceptions.
    """


class ProtocolError(AnnexError):
    """
    Base class for protocol errors
    """


class UnsupportedRequest(ProtocolError):
    """
    Must be raised when an optional request is not supported by the backend.
    """


class UnexpectedMessage(ProtocolError):
    """
    Raised when git-annex sends a message which is not expected at the moment
    """


class BackendError(AnnexError):
    """
    Must be raised by the backend when a request did not succeed.
    """


class NotLinkedError(AnnexError):
    """
    Will be raised when a Master instance is accessed without being
    linked to a Backend instance
    """


class Protocol(object):
    """
    Helper class handling the receiving part of the protocol (git-annex to
    backend) It parses the requests coming from git-annex and calls the
    respective method of the backend object.
    """

    def __init__(self, backend):
        self.backend = backend
        self.version = "VERSION 1"

    def command(self, line):
        line = line.strip()
        if not line:
            raise ProtocolError("Got empty line")
        parts = line.split(" ", 1)

        method = self.lookupMethod(parts[0])
        if method is None:
            raise UnsupportedRequest(f'Unknown request {line!r}')

        try:
            if len(parts) == 1:
                reply = method()
            else:
                reply = method(parts[1])
        except TypeError as e:
            raise SyntaxError(e)
        else:
            return reply

    def lookupMethod(self, command):
        return getattr(self, 'do_' + command.upper(), None)

    def do_GETVERSION(self):
        return self.version

    def do_CANVERIFY(self):
        return 'CANVERIFY-YES' if self.backend.can_verify() else 'CANVERIFY-NO'

    def do_ISSTABLE(self):
        return 'ISSTABLE-YES' if self.backend.is_stable() else 'ISSTABLE-NO'

    def do_ISCRYPTOGRAPHICALLYSECURE(self):
        return 'ISCRYPTOGRAPHICALLYSECURE-YES' \
            if self.backend.is_cryptographically_secure() \
            else 'ISCRYPTOGRAPHICALLYSECURE-NO'

    def do_GENKEY(self, *arg):
        try:
            key = self.backend.gen_key(arg[0])
            return f'GENKEY-SUCCESS {key}'
        except BackendError as e:
            return f'GENKEY-FAILURE {str(e)}'

    def do_VERIFYKEYCONTENT(self, *arg):
        try:
            success = self.backend.verify_content(*arg[0].split(" ", 1))
        except BackendError:
            success = False
        return 'VERIFYKEYCONTENT-SUCCESS' if success \
            else 'VERIFYKEYCONTENT-FAILURE'

    def do_ERROR(self, message):
        self.backend.error(message)


class Master(object):
    """
    Metaclass for backends.

    Attributes
    ----------
    input : io.TextIOBase
        Where to listen for git-annex request messages.
        Default: sys.stdin
    output : io.TextIOBase
        Where to send replies and backend messages
        Default: sys.stdout
    backend : Backend
        A class implementing the Backend interface to which this master
        is linked.
    """

    def __init__(self, output=sys.stdout):
        """
        Initialize the Master with an output.

        Parameters
        ----------
        output : io.TextIOBase
            Where to send replies and backend messages
            Default: sys.stdout
        """
        self.output = output

    def LinkBackend(self, backend):
        """
        Link the Master to a backend. This must be done before calling Listen()

        Parameters
        ----------
        backend : Backend
            A class implementing Backend interface to which this master
            will be linked.
        """
        self.backend = backend
        self.protocol = Protocol(backend)

    def Listen(self, input=sys.stdin):
        """
        Listen on `input` for messages from git annex.

        Parameters
        ----------
        input : io.TextIOBase
            Where to listen for git-annex request messages.
            Default: sys.stdin

        Raises
        ----------
        NotLinkedError
            If there is no backend linked to this master.
        """
        if not (hasattr(self, 'backend') and hasattr(self, 'protocol')):
            raise NotLinkedError("Please execute LinkBackend(backend) first.")

        self.input = input
        while True:
            # due to a bug in python 2 we can't use an iterator here: https://bugs.python.org/issue1633941
            line = self.input.readline()
            if not line:
                break
            line = line.rstrip()
            try:
                reply = self.protocol.command(line)
                if reply:
                    self._send(reply)
            except UnsupportedRequest as e:
                self.debug(str(e))
                self._send("UNSUPPORTED-REQUEST")
            except Exception as e:
                for line in traceback.format_exc().splitlines():
                    self.debug(line)
                self.error(e)
                raise SystemExit

    def debug(self, *args):
        """
        Tells git-annex to display the message if --debug is enabled.

        Parameters
        ----------
        message : str
            The message to be displayed to the user
        """

        self._send("DEBUG", *args)

    def error(self, *args):
        """
        Generic error. Can be sent at any time if things get too messed up to continue.
        When possible, raise a BackendError inside the respective functions.
        The backend program should exit after sending this, as git-annex will
        not talk to it any further.

        Parameters
        ----------
        error_msg : str
            The error message to be sent to git-annex
        """
        self._send("ERROR", *args)

    def progress(self, progress):
        """
        Indicates the current progress of the transfer (in bytes). May be repeated 
        any number of times during the transfer process, but it's wasteful to update
        the progress until at least another 1% of the file has been sent.
        This is highly recommended for ``*_store()``. (It is optional but good for
        ``*_retrieve()``.)

        Parameters
        ----------
        progress : int
            The current progress of the transfer in bytes.
        """
        self._send("PROGRESS {progress}".format(progress=int(progress)))

    def _send(self, *args, **kwargs):
        print(*args, file=self.output, **kwargs)
        self.output.flush()



from __future__ import annotations

import logging
from pathlib import (
    Path,
    PurePosixPath,
)
from queue import Queue
from shlex import quote as posix_quote
from typing import (
    BinaryIO,
    Callable,
)

from .common import DownloadResponseGenerator
from ..shell import (
    ExecutionResult,
    ShellCommandExecutor,
    create_result,
)
from datalad_next.consts import COPY_BUFSIZE


__all__ = [
    'DownloadResponseGenerator',
    'DownloadResponseGeneratorPosix',
    'upload',
    'download',
    'delete',
]


lgr = logging.getLogger("datalad.ext.next.shell.operations")


class DownloadResponseGeneratorPosix(DownloadResponseGenerator):
    """A response generator for efficient download commands from Linux systems"""

    def get_final_command(self, remote_file_name: bytes) -> bytes:
        """Return a final command list for the download of ``remote_file_name``

        The POSIX version for download response generators.

        This method is usually only called by
        :meth:`ShellCommandExecutor.__call__`.

        Parameters
        ----------
        remote_file_name : bytes
            The name of the file that should be downloaded. If the file name
            contains special character, e.g. space or ``$``, it must be
            quoted for a POSIX shell, for example with ``shlex.quote``.

        Returns
        -------
        bytes
            The final command that will be executed in the persistent shell
            in order to start the download in the connected shell.
        """
        command = b"""
            test -r {remote_file_name}
            if [ $? -eq 0 ]; then
                LC_ALL=C ls -dln -- {remote_file_name} | awk '{print $5; exit}'
                cat {remote_file_name}
                echo $?
            else
                echo -1;
            fi
        """.replace(b'{remote_file_name}', remote_file_name)
        return command


def upload(
    shell: ShellCommandExecutor,
    local_path: Path,
    remote_path: PurePosixPath,
    *,
    progress_callback: Callable[[int, int], None] | None = None,
    check: bool = False,
) -> ExecutionResult:
    """Upload a local file to a named file in the connected shell

    This function uploads a file to the connected shell ``shell``. It uses
    ``head`` to limit the number of bytes that the remote shell will read.
    This ensures that the upload is terminated.

    The requirements for upload are:
    - The connected shell must be a POSIX shell.
    - ``head`` must be installed in the remote shell.

    Parameters
    ----------
    shell : ShellCommandExecutor
        The shell that should be used to upload the file.
    local_path : Path
        The path of the file that should be uploaded.
    remote_path : PurePosixPath
        The path of the file on the connected shell that will contain the
        uploaded content.
    progress_callback : callable[[int, int], None], optional, default: None
        If given, the callback is called with the number of bytes that have
        been sent and the total number of bytes that should be sent.
    check : bool, optional, default: False
        If ``True``, raise a :class:`CommandError` if the remote operation does
        not exit with a ``0`` as return code.

    Returns
    -------
    ExecutionResult
        The result of the upload operation.

    Raises
    -------
    CommandError:
        If the remote operation does not exit with a ``0`` as return code, and
        ``check`` is ``True``, a :class:`CommandError` is raised. It will
        contain the exit code and the last (up to ``chunk_size`` (defined by the
        ``chunk_size`` keyword argument to :func:`shell`)) bytes of stderr
        output.
    """

    def signaling_read(
            file: BinaryIO,
            size: int,
            queue: Queue,
            *,
            chunk_size: int = COPY_BUFSIZE
    ):
        """iterator that reads from a file and signals EOF via a queue

        This iterator is used to prevent the situation where a file that
        should be uploaded is completely read and uploaded, but the final
        EOF-triggering `read()` call has not yet been made. In this case it can
        happen that the server provides an answer. If the answer is interpreted
        as indicator for a completed operation, the calling code assumes
        that it can close all file handles associated with the operation. This
        can lead to the final `read()` call being performed on a closed file,
        which would raise a `ValueError`. To prevent this, ``signaling_read``
        signals the end of the read-operation, i.e. an EOF was read, by
        enqueuing ``Ǹone`` into the signaling queue. The caller can wait for
        that event to ensure that the read operation is really done.
        """
        processed = 0
        while True:
            data = file.read(chunk_size)
            if data == b"":
                break
            yield data
            processed += len(data)
            if progress_callback is not None:
                progress_callback(processed, size)
        queue.put(None)

    # The following command line ensures that content that we send to the shell
    # either goes to the destination file or into `/dev/null`, but not into the
    # stdin of the shell. In the latter case it would be interpreted as the
    # next command, and that might be bad, e.g. if the uploaded content was
    # `rm -rf $HOME`.
    file_size = local_path.stat().st_size
    cmd_line = (
        f'head -c {file_size} > {posix_quote(str(remote_path))}'
        f"|| (head -c {file_size} > /dev/null; test 1 == 2)"
    )
    with local_path.open("rb") as local_file:
        # We use the `signaling_read` iterator to deal with the situation where
        # the content of a file that should be uploaded is completely read and
        # uploaded, but the final, EOF-triggering, `read()` call has not yet been
        # made. In this case it can happen that the server provides an answer,
        # and we leave the context, thereby closing the file. When the
        # `iterable_subprocess.<locals>.input_to`-thread then tries to read
        # from the file, a `ValueError` would be raised. This exception would
        # in turn lead to the closing of stdin of the `shell`-subprocess and
        # render it unusable.`signaling_read` allows us to wait for a completed
        # read, including the EOF reading.
        signal_queue: Queue = Queue()
        result = shell(
            cmd_line, stdin=signaling_read(local_file, file_size, signal_queue)
        )
        signal_queue.get()
    if check:
        result.to_exception(cmd_line, 'upload failed')
    return result


def download(
    shell: ShellCommandExecutor,
    remote_path: PurePosixPath,
    local_path: Path,
    *,
    progress_callback: Callable[[int, int], None] | None = None,
    response_generator_class: type[
        DownloadResponseGenerator
    ] = DownloadResponseGeneratorPosix,
    check: bool = False,
) -> ExecutionResult:
    """Download a file from the connected shell

    This method downloads a file from the connected shell.

    The requirements for download via instances of class
    :class:`DownloadResponseGeneratorPosix` are:
    - The connected shell must support `ls -dln`.
    - The connected shell must support `echo -e`.
    - The connected shell must support `awk`.
    - The connected shell must support `cat`.

    Parameters
    ----------
    shell: ShellCommandExecutor
        The shell from which a file should be downloaded.
    remote_path : PurePosixPath
        The path of the file on the connected shell that should be
        downloaded.
    local_path : Path
        The path of the local file that will contain the downloaded content.
    progress_callback : callable[[int, int], None], optional, default: None
        If given, the callback is called with the number of bytes that have
        been received and the total number of bytes that should be received.
    response_generator_class : type[DownloadResponseGenerator], optional, default: DownloadResponseGeneratorPosix
        The response generator that should be used to handle the download
        output. It must be a subclass of :class:`DownloadResponseGenerator`.
        The default works if the connected shell runs on a Unix-like system that
        provides `ls -dln`, `cat`, `echo`, and `awk`, e.g. ``Linux`` or ``OSX``.
    check : bool, optional, default: False
        If ``True``, raise a :class:`CommandError` if the remote operation does
        not exit with a ``0`` as return code.

    Returns
    -------
    ExecutionResult
        The result of the download operation.

    Raises
    -------
    CommandError:
        If the remote operation does not exit with a ``0`` as return code, and
        ``check`` is ``True``, a :class:`CommandError` is raised. It will
        contain the exit code and the last (up to ``chunk_size`` (defined by the
        ``chunk_size`` keyword argument to :func:`shell`)) bytes of stderr
        output.
    """
    command = posix_quote(str(remote_path)).encode()
    response_generator = response_generator_class(shell.stdout)
    result_generator = shell.start(
        command,
        response_generator=response_generator,
    )
    with local_path.open("wb") as local_file:
        processed = 0
        for chunk in result_generator:
            local_file.write(chunk)
            processed += len(chunk)
            if progress_callback is not None:
                progress_callback(processed, response_generator.length)

    stderr = b''.join(result_generator.stderr_deque)
    result_generator.stderr_deque.clear()
    return create_result(
        result_generator,
        command,
        stdout=b'',
        stderr=stderr,
        check=check,
        error_message='download failed',
    )


def delete(
    shell: ShellCommandExecutor,
    files: list[PurePosixPath],
    *,
    force: bool = False,
    check: bool = False,
) -> ExecutionResult:
    """Delete files on the connected shell

    The requirements for delete are:
    - The connected shell must be a POSIX shell.
    - ``rm`` must be installed in the remote shell.

    Parameters
    ----------
    shell: ShellCommandExecutor
        The shell from which a file should be downloaded.
    files : list[PurePosixPath]
        The "paths"  of the files that should be deleted.
    force : bool
        If ``True``, enforce removal, if possible. For example, the command
        could change the permissions of the files to be deleted to ensure
        their removal.
    check : bool, optional, default: False
        If ``True``, raise a :class:`CommandError` if the remote operation does
        not exit with a ``0`` as return code.

    Raises
    -------
    CommandError:
        If the remote operation does not exit with a ``0`` as return code, and
        ``check`` is ``True``, a :class:`CommandError` is raised. It will
        contain the exit code and the last (up to ``chunk_size`` (defined by the
        ``chunk_size`` keyword argument to :func:`shell`)) bytes of stderr
        output.
    """
    cmd_line = (
        "rm "
        + ("-f " if force else "")
        + " ".join(
            f"{posix_quote(str(f))}"
            for f in files
        )
    )
    result = shell(cmd_line.encode())
    if check:
        result.to_exception(cmd_line, 'delete failed')
    return result



from __future__ import annotations

from collections.abc import Callable
from functools import wraps
import warnings

__all__ = ['deprecated']


_base_tmpl = "{mod}.{func} was deprecated in version {version}. {msg}"
_kwarg_tmpl = f"argument {{kwarg!r}} of {_base_tmpl}"
_kwarg_val_tmpl = f"Use of values {{kwarg_values!r}} for {_kwarg_tmpl}"


# we must have a secret value that indicates "no value deprecation", otherwise
# we cannot tell whether `None` is deprecated or not
class _NoDeprecatedValue:
    pass


def deprecated(
        msg: str,
        *,
        version: str,
        kwarg: str | None = None,
        kwarg_values: list | _NoDeprecatedValue = _NoDeprecatedValue,
) -> Callable:
    """Annotate functions, classes, or (required) keyword-arguments
    with standardized deprecation warnings.

    Support for deprecation messages on individual keyword arguments
    is limited to calls with explicit keyword-argument use, not (implicit)
    use as a positional argument.

    Parameters
    ----------
    msg: str
      Custom message to append to a deprecation warning
    version: str
      Software version number at which the deprecation was made
    kwarg: str, optional
      Name of the particular deprecated keyword argument (instead of entire
      function/class)
    kwarg_values: list, optional
      Particular deprecated values of the specified keyword-argument
    """
    # normalize to a set(), when the set is empty, no particular value
    # was deprecated
    kwarg_values = set() \
        if kwarg_values is _NoDeprecatedValue else set(kwarg_values)

    def decorator(func):
        @wraps(func)
        def func_with_deprecation_warning(*args, **kwargs):
            # this is the layer that run for a deprecated call

            # do we have a deprecated kwargs, but it has not been used?
            # -> quick way out
            if kwarg is not None and kwarg not in kwargs.keys():
                # there is nothing to deprecate
                return func(*args, **kwargs)

            # pick the right message template
            # whole thing deprecated, or kwargs, or particular kwarg-value
            template = _base_tmpl if kwarg is None \
                else _kwarg_tmpl if not kwarg_values \
                else _kwarg_val_tmpl

            # deprecated value to compare against
            val = kwargs.get(kwarg, _NoDeprecatedValue)

            # comprehensive set of conditions when to issue deprecation
            # warning
            # - no particular kwarg is deprecated, but the whole callable
            # - no particular value is deprecated, but the whole argument
            # - given value matches any deprecated value
            # - given list/tuple/dict-keys match any deprecated value
            if (# no particular kwarg is deprecated, but the whole callable
                kwarg is None
                # no particular value is deprecated, but the whole argument
                or not kwarg_values
                # given value matches any deprecated value
                # exluce tuple/list, because they are not hashable
                or (not isinstance(val, (list, dict))
                    and val in kwarg_values)
                # given list/tuple-item or dict-key match any deprecated value
                or (isinstance(val, (tuple, list, dict))
                    and kwarg_values.intersection(val))
            ):
                warnings.warn(
                    template.format(
                        mod=func.__module__,
                        func=func.__name__,
                        kwarg=kwarg,
                        kwarg_values=kwarg_values,
                        version=version,
                        msg=msg,
                    ),
                    DeprecationWarning,
                )
            return func(*args, **kwargs)

        return func_with_deprecation_warning

    return decorator



import pytest

from ..multihash import (
    MultiHash,
    NoOpHash,
)


def test_multihash():
    mh = MultiHash(['sha1', 'MD5'])
    mh.update(b'')
    hd = mh.get_hexdigest()
    assert len(hd) == 2
    # algorithm label preserves original casing
    assert hd['MD5'] == 'd41d8cd98f00b204e9800998ecf8427e'
    assert hd['sha1'] == 'da39a3ee5e6b4b0d3255bfef95601890afd80709'

    with pytest.raises(ValueError):
        MultiHash(['bogus'])



def test_noophash():
    mh = NoOpHash()
    mh.update(b'')
    assert mh.get_hexdigest() == {}



"""Constraints for DataLad datasets"""

from __future__ import annotations

from pathlib import (
    Path,
    PurePath,
)

from datalad_next.datasets import Dataset

from .base import (
    Constraint,
    DatasetParameter,
)
from .exceptions import NoDatasetFound


class EnsureDataset(Constraint):
    """Ensure an absent/present `Dataset` from any path or Dataset instance

    Regardless of the nature of the input (`Dataset` instance or local path)
    a resulting instance (if it can be created) is optionally tested for
    absence or presence on the local file system.

    Due to the particular nature of the `Dataset` class (the same instance
    is used for a unique path), this constraint returns a `DatasetParameter`
    rather than a `Dataset` directly. Consuming commands can discover
    the original parameter value via its `original` property, and access a
    `Dataset` instance via its `ds` property.

    In addition to any value representing an explicit path, this constraint
    also recognizes the special value `None`. This instructs the implementation
    to find a dataset that contains the process working directory (PWD).
    Such a dataset need not have its root at PWD, but could be located in
    any parent directory too. If no such dataset can be found, PWD is used
    directly. Tests for ``installed`` are performed in the same way as with
    an explicit dataset location argument. If `None` is given and
    ``installed=True``, but no dataset is found, an exception is raised
    (this is the behavior of the ``required_dataset()`` function in
    the DataLad core package). With ``installed=False`` no exception is
    raised and a dataset instances matching PWD is returned.
    """
    def __init__(self,
                 installed: bool | None = None,
                 purpose: str | None = None,
                 require_id: bool | None = None):
        """
        Parameters
        ----------
        installed: bool, optional
          If given, a dataset will be verified to be installed or not.
          Otherwise the installation-state will not be inspected.
        purpose: str, optional
          If given, will be used in generated error messages to communicate
          why a dataset is required (to exist)
        idcheck: bool, option
          If given, performs an additional check whether the dataset has a
          valid dataset ID.
        """
        self._installed = installed
        self._purpose = purpose
        self._require_id = require_id
        super().__init__()

    def __call__(self, value) -> DatasetParameter:
        # good-enough test to recognize a dataset instance cheaply
        if hasattr(value, 'repo') and hasattr(value, 'pathobj'):
            ds = value
        # anticipate what require_dataset() could handle and fail if we got
        # something else
        elif not isinstance(value, (str, PurePath, type(None))):
            self.raise_for(
                value, "cannot create Dataset from {type}", type=type(value)
            )
        else:
            ds = self._require_dataset(value)
        assert ds
        if self._installed is not None:
            is_installed = ds.is_installed()
            if self._installed is False and is_installed:
                self.raise_for(ds, 'already exists locally')
            if self._installed and not is_installed:
                self.raise_for(ds, 'not installed')
        if self._require_id and not ds.id:
            self.raise_for(ds, 'does not have a valid datalad-id')
        return DatasetParameter(value, ds)

    def short_description(self) -> str:
        return "(path to) {}Dataset".format(
            'an existing ' if self._installed is True
            else 'a non-existing ' if self._installed is False else 'a ')


    def _require_dataset(self, value):
        from datalad.distribution.dataset import require_dataset
        try:
            ds = require_dataset(
                value,
                check_installed=self._installed is True,
                purpose=self._purpose,
            )
            return ds
        except NoDatasetFound:
            # mitigation of non-uniform require_dataset() behavior.
            # with value == None it does not honor check_installed
            # https://github.com/datalad/datalad/issues/7281
            if self._installed is True:
                # if we are instructed to ensure an installed dataset
                raise
            else:
                # but otherwise go with CWD. require_dataset() did not
                # find a dataset in any parent dir either, so this is
                # the best we can do. Installation absence verification
                # will happen further down
                return Dataset(Path.cwd())



"""Credential management

.. currentmodule:: datalad_next.credman
.. autosummary::
   :toctree: generated

   CredentialManager
   verify_property_names
"""
from .manager import (
    CredentialManager,
    verify_property_names,
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Credential management and query"""

# allow for |-type UnionType declarations
from __future__ import annotations

__docformat__ = 'restructuredtext'

__all__ = ['CredentialManager']

from collections.abc import Set
from datetime import datetime
import logging
import re

from typing import (
    Dict,
    List,
    Tuple,
)

import datalad
from datalad_next.config import ConfigManager
from datalad_next.exceptions import (
    CapturedException,
    CommandError,
)
from datalad_next.uis import ui_switcher as ui

lgr = logging.getLogger('datalad.credman')


class CredentialManager(object):
    """Facility to get, set, remove and query credentials.

    A credential in this context is a set of properties (key-value pairs)
    associated with exactly one secret.

    At present, the only backend for secret storage is the Python keyring
    package, as interfaced via a custom DataLad wrapper. Store for credential
    properties is implemented using DataLad's (i.e. Git's) configuration
    system. All properties are stored in the `global` (i.e., user) scope
    under configuration items following the pattern::

      datalad.credential.<name>.<property>

    where ``<name>`` is a credential name/identifier, and ``<property>`` is an
    arbitrarily named credential property, whose name must follow the
    git-config syntax for variable names (case-insensitive, only alphanumeric
    characters and ``-``, and must start with an alphabetic character).

    Create a ``CredentialManager`` instance is fast, virtually no
    initialization needs to be performed. All internal properties are lazily
    evaluated.  This facilitates usage in code where it is difficult to
    incorporate a long-lived central instance.

    API

    With one exception, all parameter names of methods in the core API
    outside ``**kwargs`` must have a ``_`` prefix that distinguishes credential
    properties from method parameters. The one exception is the ``name``
    parameter, which is used as a primary identifier (albeit being optional
    for some operations).

    The ``obtain()`` method is provided as an additional convenience, and
    implements a standard workflow for obtaining a credential in a wide variety
    of scenarios (credential name, credential properties, secret either
    respectively already known or yet unknown).
    """
    valid_property_names_regex = re.compile(r'[a-z0-9]+[a-z0-9-]*$')

    secret_names = {
        'user_password': 'password',
    }

    def __init__(self, cfg: ConfigManager | None = None):
        """

        Parameters
        ----------
        cfg: ConfigManager, optional
          If given, all configuration queries are performed using this
          ``ConfigManager`` instance. Otherwise ``datalad.cfg`` is used.
        """
        self.__cfg = cfg
        self.__cred_types = None
        self.__keyring = None

    # main API
    #

    #
    # Design remark:
    # This is the keyhole through which any retrieval-related processing goes
    # (get, query, obtain). Any normalization performed in here will
    # automatically also effect these other operations.
    #
    def get(self, name=None, *, _prompt=None, _type_hint=None, **kwargs):
        """Get properties and secret of a credential.

        This is a read-only method that never modifies information stored
        on a credential in any backend.

        Credential property lookup is supported via a number approaches.  When
        providing ``name``, all existing corresponding configuration items are
        found and reported, and an existing secret is retrieved from name-based
        secret backends (presently ``keyring``). When providing a ``type``
        property or a ``_type_hint`` the lookup of additional properties in the
        keyring-backend is enabled, using predefined property name lists
        for a number of known credential types.

        For all given property keys that have no value assigned after the
        initial lookup, manual/interactive entry is attempted, whenever
        a custom ``_prompt`` was provided. This include requesting a secret.
        If manually entered information is contained in the return credential
        record, the record contains an additional ``_edited`` property with a
        value of ``True``.

        If no secret is known after lookup and a potential manual data entry,
        a plain ``None`` is returned instead of a full credential record.

        Parameters
        ----------
        name: str, optional
          Name of the credential to be retrieved
        _prompt: str or None
          Instructions for credential entry to be displayed when missing
          properties are encountered. If ``None``, manual entry is disabled.
        _type_hint: str or None
          In case no ``type`` property is included in ``kwargs``, this parameter
          is used to determine a credential type, to possibly enable further
          lookup/entry of additional properties for a known credential type
        **kwargs:
          Credential property name/value pairs to overwrite/amend potentially
          existing properties. For any property with a value
          of ``None``, manual data entry will be performed, unless a value
          could be retrieved on lookup, or prompting was not enabled.

        Returns
        -------
        dict or None
          Return ``None``, if no secret for the credential was found or entered.
          Otherwise returns the complete credential record, comprising all
          properties and the secret. An additional ``_edited`` key with a
          value of ``True`` is added whenever the returned record contains
          manually entered information.

        Raises
        ------
        ValueError
          When the method is called without any information that could be
          used to identify a credential
        """
        if name is None and _type_hint is None and not kwargs:
            # there is no chance that this could work
            raise ValueError(
                'CredentialManager.get() called without any identifying '
                'information')
        if name is None:
            # there is no chance we can retrieve any stored properties
            # but we can prompt for some below
            cred = {}
        else:
            # if we have a chance to query for stored legacy credentials
            # we do this first to have the more modern parts of the
            # system overwrite them reliably
            cred = self._get_legacy_credential_from_keyring(
                name,
                type_hint=kwargs.get('type', _type_hint),
            ) or {}

            # and now take whatever we got from the legacy store and update
            # it from what we have in the config
            cred_update = self._get_credential_from_cfg(name)
            if set(cred_update.keys()) >= set(
                    k for k in cred.keys() if k != '_from_backend'):
                # we are overwriting all info from a possible legacy credential
                # take marker off
                cred.pop('_from_backend', None)
            cred.update(cred_update)

        # we merge existing credential properties with overrides.
        # we are only adding 'enter-please' markers (i.e. None) for properties
        # that have no known value yet
        cred.update(**{k: v for k, v in kwargs.items()
                       if v is not None or k not in cred})

        # final word on the credential type, if there is any type info at all
        # `cred` will have a `type` property after this
        self._assign_credential_type(cred, _type_hint)

        # determine what is missing and possibly prompt for it
        self._complete_credential_props(name, cred, _prompt)

        if not cred.get('secret'):
            # no secret, no credential
            if any(not p.startswith('_') for p in cred):
                lgr.debug(
                    'Not reporting on credential fragment '
                    '(name=%r) with no secret: %r',
                    name, cred,
                )
            return

        return cred

    def set(self,
            name,
            *,
            _lastused=False,
            _suggested_name=None,
            _context=None,
            **kwargs):
        """Set credential properties and secret

        Presently, all supported backends require the specification of
        a credential ``name`` for storage. This may change in the future,
        when support for alternative backends is added, at which point
        the ``name`` parameter would become optional.

        All properties provided as `kwargs` with keys not starting with `_` and
        with values that are not ``None`` will be stored. If ``kwargs`` do not
        contain a ``secret`` specification, manual entry will be attempted. The
        associated prompt with be either the name of the ``secret`` field of a
        known credential (as identified via a ``type`` property), or the label
        ``'secret'``.

        All properties with an associated value of ``None`` will be removed
        (unset).

        Parameters
        ----------
        name: str or None
          Credential name. If None, the name will be prompted for and setting
          the credential is skipped if no name is provided.
        _lastused: bool, optional
          If set, automatically add an additional credential property
          ``'last-used'`` with the current timestamp in ISO 8601 format.
        _suggested_name: str, optional
          If `name` is None, this name (if given) is presented as a default
          suggestion that can be accepted without having to enter it manually.
          If this name suggestion conflicts with an existing credential, it
          is ignored and not presented as a suggestion.
        _context: str, optional
          If given, will be included in the prompt for a missing credential
          name to provide context for a user. It should be  written to fit
          into a parenthical statement after "Enter a name to save the
          credential (...)", e.g. "for download from <URL>".
        **kwargs:
          Any number of credential property key/value pairs to set (update),
          or remove. With one exception, values of ``None`` indicate removal
          of a property from a credential. However, ``secret=None`` does not
          lead to the removal of a credential's secret, because it would
          result in an incomplete credential. Instead, it will cause a
          credential's effective ``secret`` property to be written to the
          secret store. The effective secret might come from other sources,
          such as particular configuration scopes or environment variables
          (i.e., matching the ``datalad.credential.<name>.secret``
          configuration item.
          Properties whose names start with an underscore are automatically
          removed prior storage.

        Returns
        -------
        dict or None
          key/values of all modified credential properties with respect
          to their previously recorded values. None is returned in case
          a user did not enter a missing credential name. If a user entered
          a credential name, it is included in the returned dictionary
          under the 'name' key.

        Raises
        ------
        RuntimeError
          This exception is raised whenever a property cannot be removed
          successfully. Likely cause is that it is defined in a configuration
          scope or backend for which write-access is not supported.
        ValueError
          When property names in kwargs are not syntax-compliant.
        """
        updated = {}

        if not name:
            known_credentials = self._get_known_credential_names()
            if _suggested_name in known_credentials:
                # ignore name suggestion, conflicts with existing credential
                _suggested_name = None
            prompt = 'Enter a name to save the credential'
            if _context:
                prompt = f'{prompt} ({_context})'
            prompt = f"{prompt} securely for future reuse, " \
                     "or 'skip' to not save the credential"
            if _suggested_name:
                prompt = f'{prompt}, or leave empty to accept the name ' \
                         f'{_suggested_name!r}'
            while not name:
                entered = self._ask_property('name', prompt=prompt)
                if entered == 'skip':
                    return
                elif entered is None:
                    # the user was no able to enter a value (non-interactive
                    # session). we raise to not let this go unnoticed.
                    raise ValueError('no credential name provided for setting')
                elif not entered:
                    if not _suggested_name:
                        ui.message('Cannot proceed without a credential name')
                        continue
                    # otherwise take the default
                    entered = _suggested_name
                if entered in known_credentials:
                    ui.message(
                        f'A credential with the name {entered!r} already '
                        'exists, please provide a different name.')
                else:
                    name = entered
            updated['name'] = name

        # we strip internal properties, such as '_edited' automatically
        # forcing each caller to to this by hand is kinda pointless, if
        # they can never be stored anyway, and e.g. a previous `get()`
        # would include one for any credentials that was manually entered
        kwargs = self._strip_internal_properties(kwargs)
        # check syntax for the rest
        verify_property_names(kwargs)
        # retrieve the previous credential state, so we can merge with the
        # incremental changes, and report on effective updates
        prev_cred = self.get(
            name=name,
            # we never want any manual interaction at this point
            _prompt=None,
            # if we know the type, hence we can do a query for legacy secrets
            # and properties. This will migrate them to the new setup
            # over time
            _type_hint=kwargs.get('type'),
        )
        # merge incoming with existing properties to create an updated
        # credential
        if prev_cred:
            prev_cred = self._strip_internal_properties(prev_cred)
            cred = dict(prev_cred, **kwargs)
        else:
            cred = dict(kwargs)
        # update last-used, if requested
        if _lastused:
            cred['last-used'] = datetime.now().isoformat()

        # remove props
        #
        remove_props = [
            k for k, v in cred.items()
            # can we really know that no 'secret' field was deposited
            # in the config backend? MIH does not think so. However,
            # secret=None has special semantics (update secret store
            # from config), hence we cannot use it to perform removal
            # of secrets from config here.
            # MIH did not find a rational for this setup. It was already
            # part of the original implementation. At least this is
            # documented in the `kwargs` docstring now.
            if v is None and k != 'secret']
        self._unset_credprops_anyscope(name, remove_props)
        updated.update(**{k: None for k in remove_props})

        # set non-secret props
        #
        set_props = {
            k: v for k, v in cred.items()
            if v is not None and k != 'secret'
        }
        for k, v in set_props.items():
            var = _get_cred_cfg_var(name, k)
            if self._cfg.get(var) == v:
                # desired value already exists, we are not
                # storing again to preserve the scope it
                # was defined in
                continue
            # we always write to the global scope (ie. user config)
            # credentials are typically a personal, not a repository
            # specific entity -- likewise secrets go into a personal
            # not repository-specific store
            # for custom needs users can directly set the respective
            # config
            self._cfg.set(var, v, scope='global', force=True, reload=False)
            updated[k] = v
        if set_props:
            self._cfg.reload()

        # set secret
        #
        # we aim to update the secret in the store, hence we must
        # query for a previous setting in order to reliably report
        # updates
        prev_secret = self._get_secret(prev_cred) if prev_cred else None
        if 'secret' not in cred:
            # we have no removal directive, reuse previous secret
            cred['secret'] = prev_secret
        if cred.get('secret') is None:
            # we want to reset the secret, consider active config
            # (which was already queried when retrieving the previous
            # credential above)
            cred['secret'] = prev_secret

        if cred.get('secret') is None:
            # we have no secret specified or in the store already: ask
            # typically we would end up here with an explicit attempt
            # to set a credential in a context that is known to an
            # interactive user, hence the messaging here can be simple
            cred['secret'] = self._ask_secret(
                CredentialManager.secret_names.get(cred.get('type'), 'secret'))
        # at this point we will have a secret. it could be from ENV
        # or provided, or entered. we always want to put it in the
        # store
        # we never ever write a secret to any other field-name than
        # 'secret'
        if cred['secret'] != self._keyring.get(name, 'secret'):
            # only report updated if actually different from before.
            # and "before" is what was in the secret store, because
            # we will write to it next. A secret could have been
            # provided via an ENV var, hence even with no change from
            # `prev_cred` there could be a change in the secret store
            updated['secret'] = cred['secret']
        # TODO make sure that there actually is a secret that is written
        # and not None
        self._keyring.set(name, 'secret', cred['secret'])
        return updated

    def remove(self, name, *, type_hint=None):
        """Remove a credential, including all properties and secret

        Presently, all supported backends require the specification of
        a credential ``name`` for lookup. This may change in the future,
        when support for alternative backends is added, at which point
        the ``name`` parameter would become optional, and additional
        parameters would be added.

        Returns
        -------
        bool
          True if a credential was removed, and False if not (because
          no respective credential was found).

        Raises
        ------
        RuntimeError
          This exception is raised whenever a property cannot be removed
          successfully. Likely cause is that it is defined in a configuration
          scope or backend for which write-access is not supported.
        """
        # prefix for all config variables of this credential
        prefix = _get_cred_cfg_var(name, '')

        to_remove = [
            k[len(prefix):] for k in self._cfg.keys()
            if k.startswith(prefix)
        ]
        removed = False
        if to_remove:
            self._unset_credprops_anyscope(name, to_remove)
            removed = True

        # delete the secret from the keystore, if there is any
        def del_field(name, field):
            global removed
            try:
                self._keyring.delete(name, field)
                removed = True
            except Exception as e:
                if self._keyring.get(name, field) is None:
                    # whatever it was, the target is reached
                    CapturedException(e)
                else:
                    # we could not delete the field
                    raise  # pragma: nocover

        del_field(name, 'secret')
        if type_hint:
            # remove legacy records too
            for field in self._cred_types.get(
                    type_hint, {}).get('fields', []):
                del_field(name, field)
        return removed

    def query_(self, **kwargs):
        """Query for all (matching) credentials.

        Credentials are yielded in no particular order.

        This method cannot find credentials for which only a secret
        was deposited in the keyring.

        This method does support lookup of credentials defined in DataLad's
        "provider" configurations.

        Parameters
        ----------
        **kwargs
          If not given, any found credential is yielded. Otherwise,
          any credential must match all property name/value
          pairs

        Yields
        ------
        tuple(str, dict)
          The first element in the tuple is the credential name, the second
          element is the credential record as returned by ``get()`` for any
          matching credential.
        """
        done = set()
        known_credentials = set((n, None) for n in self._get_known_credential_names())
        from itertools import chain
        for name, legacy_type_hint in chain(
                _yield_legacy_credential_types(),
                known_credentials):
            done.add(name)
            cred = self.get(name, _prompt=None, _type_hint=legacy_type_hint)
            if not cred and legacy_type_hint:
                # this legacy-type credential is not set. We still want to
                # report on it, because it is the only way for users that
                # discover these predefined credential "slots"
                cred = dict(type=legacy_type_hint)
            if legacy_type_hint is not None:
                # leading underscore to distinguish this internal marker from
                # an actual credential property.
                # the credentials command will then also treat it as such
                cred['_from_backend'] = 'legacy'
            if not cred:
                # no info on such a credential, not even legacy info
                # ignore
                continue
            if not kwargs:
                yield (name, cred)
            else:
                if all(cred.get(k) == v for k, v in kwargs.items()):
                    yield (name, cred)
                else:
                    continue

    def query(self, *, _sortby=None, _reverse=True, **kwargs):
        """Query for all (matching) credentials, sorted by a property

        This method is a companion of ``query_()``, and the same limitations
        regarding credential discovery apply.

        In contrast to ``query_()``, this method return a list instead of
        yielding credentials one by one. This returned list is optionally
        sorted.

        Parameters
        ----------
        _sortby: str, optional
          Name of a credential property to provide a value to sort by.
          Credentials that do not carry the specified property always
          sort last, regardless of sort order.
        _reverse: bool, optional
          Flag whether to sort ascending or descending when sorting.
          By default credentials are return in descending property
          value order. This flag does not impact the fact that credentials
          without the property to sort by always sort last.
        **kwargs
          Pass on as-is to ``query_()``

        Returns
        -------
        list(str, dict)
          Each item is a 2-tuple. The first element in each tuple is the
          credential name, the second element is the credential record
          as returned by ``get()`` for any matching credential.
        """
        matches = self.query_(**kwargs)
        if _sortby is None:
            return list(matches)

        # this makes sure that any credential that does not have the
        # sort-by property name sorts to the very end of the list
        # regardless of whether the sorting is ascending or descending
        def get_sort_key(x):
            # x is a tuple as returned by query_()
            prop_indicator = _sortby in x[1]
            if not _reverse:
                prop_indicator = not prop_indicator
            return (prop_indicator, x[1].get(_sortby))

        return sorted(matches, key=get_sort_key, reverse=_reverse)

    def obtain(self,
               name: str | None = None,
               *,
               prompt: str | None = None,
               type_hint: str | None = None,
               query_props: Dict | None = None,
               expected_props: List | Tuple | None = None):
        """Obtain a credential by query or prompt (if needed)

        This convenience method implements a standard workflow to obtain a
        credential.  It supports credential selection by credential
        name/identifier, and falls back onto querying for a credential matching
        a set of specified properties (as key-value mappings). If no suitable
        credential is known, a user is prompted to enter one interactively (if
        possible in the current session).

        If a credential was entered manually, any given ``type_hint`` will
        be included as a ``type`` property of the returned credential, and
        the returned credential has an ``_edited=True`` property.
        Likewise, any ``realm`` property included in the ``query_props``
        is included in the returned credential in this case.

        If desired, a credential workflow can be completed, after a credential
        was found to be valid/working, by storing or updating it in the
        credential store::

            cm = CredentialManager()
            cname, cprops = cm.obtain(...)
            # verify credential is working
            ...
            # set/update
            cm.set(cname, _lastused=True, **cprops)

        In the code sketch above, if ``cname`` is ``None`` (as it will be for
        a newly entered credential, ``set()`` will prompt for a name to store
        the credential under, and will offer a user the choice to skip storing
        a credential. For any previously known credential, the ``last-used``
        property will be updated to enable preferred selection in future
        credential discovery attempts via ``obtain()``.

        Examples
        --------

        Minimal call to get a credential entered (manually)::

          credman.obtain(type_hint='token', prompt='Credential please!')

        Without a prompt text no interaction is attempted, and without a type
        hint it is unknown what (and how much) to enter.

        Minimal call to retrieve a credential by its identifier::

          credman.obtain('my-github-token')

        Minimal call to retrieve the last-used credential for a particular
        authentication "realm". In this case "realm" is a property that was
        previously set to match a particular service/location, and is now used
        to match credentials against::

          credman.obtain(query_props={'realm': 'mysecretlair'})

        Parameters
        ----------
        name: str, optional
          Name of the credential to be retrieved
        prompt: str, optional
          Passed to ``CredentialManager.get()`` if a credential name was
          provided, or no suitable credential could be found by querying.
        type_hint: str, optional
          In case no ``type`` property is included in ``query_props``,
          this parameter is passed to ``CredentialManager.get()``.
        query_props: dict, optional
          Credential property to be used for querying for a suitable
          credential. When multiple credentials match a query, the last-used
          credential is selected.
        expected_props: list or tuple, optional
          When specified, a credential will be inspected to contain properties
          matching all listed property names, or a ``ValueError`` will be
          raised.

        Returns
        -------
        (str, dict)
          Credential name (possibly different from the input, when a credential
          was discovered based on properties), and credential properties.

        Raises
        ------
        ValueError
          Raised when no matching credential could be found and none was
          entered. Also raised, when a credential selected from a query result
          or a manually entered one is missing any of the properties with a
          name given in ``expected_props``.
        """
        cred = None
        if not name:
            if query_props:
                creds = self.query(_sortby='last-used', **(query_props or {}))
                if creds:
                    name, cred = creds[0]

        if not cred:
            kwargs = dict(
                # name could be none
                name=name,
                _prompt=prompt,
                _type_hint=type_hint,
            )
            try:
                cred = self.get(**kwargs)
                # check if we know the realm, if so include in the credential,
                realm = (query_props or {}).get('realm')
                if realm:
                    cred['realm'] = realm
                if name is None and type_hint:
                    # we are not expecting to retrieve a particular credential.
                    # make the type hint the actual type of the credential
                    cred['type'] = type_hint
            except Exception as e:
                lgr.debug('Obtaining credential failed: %s', e)

        if not cred:
            raise ValueError('No suitable credential found or specified')
        missing_props = [
            ep for ep in (expected_props or []) if ep not in cred
        ]
        if any(missing_props):
            raise ValueError(
                'No suitable credential or specified '
                f'(missing properties: {missing_props})')
        return name, cred

    # internal helpers
    #
    def _strip_internal_properties(self, cred: Dict) -> Dict:
        return {k: v for k, v in cred.items() if not k.startswith('_')}

    def _assign_credential_type(self, cred, _type_hint=None):
        """Set 'type' property (in-place)"""
        _type_hint = cred.get('type', _type_hint)
        if _type_hint:
            cred['type'] = _type_hint
            return

        # if we get here, we don't know what type this is
        # let's derive one for a few clear-cut cases where we can be
        # reasonable sure what type a credential is
        if set(cred) == set(('token',)):
            # all we have is a token property -- very likely a token-type
            # credential. Move the token to the secret property and
            # assign the type
            cred['type'] = 'token'
            cred['secret'] = cred.pop('token')

    def _complete_credential_props(
            self, name: str,
            cred: Dict,
            prompt: str | None,
    ) -> None:
        """Determine missing credential properties, and fill them in

        What properties are missing is determined based on credential
        type info, and their values will be prompted for (if a prompt
        was provided).

        The given credential is modified in place.
        """
        cred_type = cred.get('type')
        # import the definition of expected fields from the known
        # credential types
        cred_type_def = self._cred_types.get(
            cred_type,
            dict(fields=[], secret=None))
        required_fields = cred_type_def['fields'] or []
        secret_field = cred_type_def['secret']
        # mark required fields for this credential type
        for k in required_fields:
            if k == secret_field:
                # do nothing, if this is the secret key
                continue
            if k in cred:
                # do nothing if we have an incoming value for this key already
                continue
            # otherwise make sure we prompt for the essential
            # fields
            cred[k] = None

        # - prompt for required but missing prompts
        # - retrieve a secret
        prompted = False
        entered = {}
        for k, v in cred.items():
            if k in ('secret', secret_field):
                # a secret is either held in a 'secret' field, or in a dedicated field
                # defined by the cred_type_def. Both are handled below
                # handled below
                continue
            if prompt and v is None:
                # prevent double-prompting for entering a series of properties
                # of the same credential
                v = self._ask_property(k, None if prompted else prompt)
                if v is not None:
                    prompted = True
            if v:
                entered[k] = v

        # bulk merged, cannot do in-loop above, because we iterate over items()
        cred.update(entered)

        # extract the secret, from the assembled properties or a secret store
        secret = self._get_secret(cred, name=name, secret_field=secret_field)
        if prompt and secret is None:
            secret = self._ask_secret(
                type_hint=secret_field,
                prompt=None if prompted else prompt,
            )
            if secret:
                prompted = True

        if secret:
            cred['secret'] = secret

        # report whether there were any edits to the credential record
        # (incl. being entirely new), such that consumers can decide
        # to save a credentials, once battle-tested
        if prompted:
            cred['_edited'] = True

    def _get_credential_from_cfg(self, name: str) -> Dict:
        var_prefix = _get_cred_cfg_var(name, '')
        return {
            k[len(var_prefix):]: v
            for k, v in self._cfg.items()
            if k.startswith(var_prefix)
        }

    def _get_known_credential_names(self) -> Set[str]:
        known_credentials = set(
            '.'.join(k.split('.')[2:-1]) for k in self._cfg.keys()
            if k.startswith('datalad.credential.')
        )
        return known_credentials

    def _ask_property(self, name, prompt=None):
        if not ui.is_interactive:
            lgr.debug('Cannot ask for credential property %r in non-interactive session', name)
            return

        return ui.question(name, title=prompt)

    def _ask_secret(self, type_hint=None, prompt=None):
        if not ui.is_interactive:
            lgr.debug('Cannot ask for credential secret in non-interactive session')
            return

        return ui.question(
            type_hint or 'secret',
            title=prompt,
            repeat=self._cfg.obtain(
                'datalad.credentials.repeat-secret-entry'),
            hidden=self._cfg.obtain(
                'datalad.credentials.hidden-secret-entry'),
        )

    def _unset_credprops_anyscope(self, name, keys):
        """Reloads the config after unsetting all relevant variables

        This method does not modify the keystore.
        """
        nonremoved_vars = []
        for k in keys:
            var = _get_cred_cfg_var(name, k)
            if var not in self._cfg:
                continue
            try:
                self._cfg.unset(var, scope='global', reload=False)
            except CommandError as e:
                CapturedException(e)
                try:
                    self._cfg.unset(var, scope='local', reload=False)
                except CommandError as e:
                    CapturedException(e)
                    nonremoved_vars.append(var)
        if nonremoved_vars:
            raise RuntimeError(
                f"Cannot remove configuration items {nonremoved_vars} "
                f"for credential, defined outside global or local "
                "configuration scope. Remove manually")
        self._cfg.reload()

    def _get_legacy_credential_from_keyring(
            self,
            name: str,
            type_hint: str | None,
    ) -> Dict | None:
        """With a ``type_hint`` given or determined from a known legacy
        credential, attempts to retrieve a credential
        comprised of all fields defined in ``self._cred_types``. Otherwise
        ``None`` is returned.
        """
        if not type_hint:
            # no type hint given in any form. Last chance is that
            # this is a known legacy credential.
            # doing this query is a bit expensive, but getting a
            # credential is not a high-performance procedure, and
            # the gain in convenience is substantial -- otherwise
            # users would need to somehow know what they should be
            # looking for
            type_hint = dict(_yield_legacy_credential_types()).get(name)

        if not type_hint or type_hint not in self._cred_types:
            return None

        cred = {}
        lc = self._cred_types[type_hint]
        for field in (lc['fields'] or []):
            if field == lc['secret']:
                continue
            val = self._keyring.get(name, field)
            if val:
                # legacy credentials used property names with underscores,
                # but this is no longer syntax-compliant -- fix on read
                cred[field.replace('_', '-')] = val

        if not cred:
            # there is nothing on a legacy credential with this name
            return None
        else:
            # at least some info came from the legacy backend, record that
            cred['_from_backend'] = 'legacy'
            cred['type'] = type_hint
            return cred

    def _get_secret(
            self,
            cred: Dict,
            name: str | None = None,
            secret_field: str | None = None,
    ) -> str | None:
        """Report a secret

        Either directly from the set of credential properties, or from a
        secret store.
        """
        # from literal 'secret' property
        secret = cred.get('secret')
        if secret:
            return secret
        # from secret store under 'secret' label
        if name:
            secret = self._keyring.get(name, 'secret')
            if secret:
                return secret
        # `secret_field` property
        if secret_field:
            secret = cred.get(secret_field)
            if secret:
                return secret
        # from secret store under `secret_field` label
        if name and secret_field:
            secret = self._keyring.get(name, secret_field)
            if secret:
                return secret

        # no secret found anywhere
        return None

    @property
    def _cfg(self):
        """Return a ConfigManager given to __init__() or the global datalad.cfg
        """
        if self.__cfg:
            return self.__cfg
        return datalad.cfg

    @property
    def _keyring(self):
        """Returns the DataLad keyring wrapper

        This internal property may vanish whenever changes to the supported
        backends are made.
        """
        if self.__keyring:
            return self.__keyring
        from datalad.support.keyring_ import keyring
        self.__keyring = keyring
        return keyring

    @property
    def _cred_types(self):
        """Internal property for mapping of credential type names to fields.

        Returns
        -------
        dict
          Legacy credential type name ('token', 'user_password', etc.) as keys,
          and dictionaries as values. Each of these dicts has two keys:
          'fields' (the complete list of "fields" that the credential
          comprises), and 'secret' (the name of the field that represents the
          secret. If there is no secret, the value associated with that key is
          ``None``.
        """
        # at present the credential type specifications are built from the
        # legacy credential types, but this may change at any point in the
        # future
        # here is what that was in Mar 2022
        # 'user_password': {'fields': ['user', 'password'],
        #                   'secret': 'password'},
        # 'token':  {'fields': ['token'], 'secret': 'token'},
        # 'git':    {'fields': ['user', 'password'], 'secret': 'password'}
        # 'aws-s3': {'fields': ['key_id', 'secret_id', 'session', 'expiration'],
        #            'secret': 'secret_id'},
        # 'nda-s3': {'fields': None, 'secret': None},
        # 'loris-token': {'fields': None, 'secret': None},

        if self.__cred_types:
            return self.__cred_types

        from datalad.downloaders import CREDENTIAL_TYPES
        mapping = {}
        for cname, ctype in CREDENTIAL_TYPES.items():
            secret_fields = [
                f for f in (ctype._FIELDS or {})
                if ctype._FIELDS[f].get('hidden')
            ]
            mapping[cname] = dict(
                fields=list(ctype._FIELDS.keys()) if ctype._FIELDS else None,
                secret=secret_fields[0] if secret_fields else None,
            )
        # an implementation-independent s3-style credential (with the aim to
        # also work for MinIO and Ceph)
        mapping['s3'] = dict(
            # use boto-style names, but strip "aws" prefix, and redundant
            # non-distinguishing 'key' and 'access' terms
            fields=['key', 'secret'],
            secret='secret',
        )
        self.__cred_types = mapping
        return mapping


def _yield_legacy_credential_types():
    # query is constrained by non-secrets, no constraints means report all
    # a constraint means *exact* match on all given properties
    from datalad.downloaders.providers import (
        Providers,
        CREDENTIAL_TYPES,
    )
    type_hints = {v: k for k, v in CREDENTIAL_TYPES.items()}

    # ATTN: from Providers.from_config_files() is sensitive to the PWD
    # it will only read legacy credentials from datasets whenever
    # PWD is inside a dataset
    legacy_credentials = set(
        (p.credential.name, type(p.credential))
        # without reload, no changes in files since the last call
        # will be considered. That last call might have happened
        # in datalad-core, and may have been in another directory
        for p in Providers.from_config_files(reload=True)
        if p.credential
    )
    for name, type_ in legacy_credentials:
        yield (name, type_hints.get(type_))


def verify_property_names(names):
    """Check credential property names for syntax-compliance.

    Parameters
    ----------
    names: iterable

    Raises
    ------
    ValueError
      When any non-compliant property names were found
    """
    invalid_names = [
        k for k in names
        if not CredentialManager.valid_property_names_regex.match(k)
    ]
    if invalid_names:
        raise ValueError(
            f'Unsupported property names {invalid_names}, '
            'must match git-config variable syntax (a-z0-9 and - characters)')


def _get_cred_cfg_var(name, prop):
    """Return a config variable name for a credential property

    Parameters
    ----------
    name : str
      Credential name
    prop : str
      Property name

    Returns
    -------
    str
    """
    return f'datalad.credential.{name}.{prop}'



"""Handler for operations, such as "download", on http(s):// URLs"""

# allow for |-type UnionType declarations
from __future__ import annotations

import logging
from pathlib import Path
import sys
from typing import Dict
import requests
from requests_toolbelt import user_agent

import datalad

from datalad_next.utils import (
    DataladAuth,
    parse_www_authenticate,
)
from .base import UrlOperations
from .exceptions import (
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
)

lgr = logging.getLogger('datalad.ext.next.url_operations.http')


__all__ = ['HttpUrlOperations']


class HttpUrlOperations(UrlOperations):
    """Handler for operations on `http(s)://` URLs

    This handler is built on the `requests` package. For authentication, it
    employes :class:`datalad_next.utils.requests_auth.DataladAuth`, an adaptor
    that consults the DataLad credential system in order to fulfill HTTP
    authentication challenges.
    """

    def __init__(self, cfg=None, headers: Dict | None = None):
        """
        Parameters
        ----------
        cfg: ConfigManager, optional
          A config manager instance that is consulted for any configuration
          filesystem configuration individual handlers may support.
        headers: dict, optional
          Additional or alternative headers to add to a request. The default
          headers contain a ``user-agent`` declaration. Any headers provided
          here override corresponding defaults.
        """
        super().__init__(cfg=cfg)
        self._headers = {
            'user-agent': user_agent('datalad', datalad.__version__),
        }
        if headers:
            self._headers.update(headers)

    def get_headers(self, headers: Dict | None = None) -> Dict:
        # start with the default
        hdrs = dict(self._headers)
        if headers is not None:
            hdrs.update(headers)
        return hdrs

    def _check_status(self, response, url):
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            # wrap this into the datalad-standard, but keep the
            # original exception linked
            if e.response.status_code == 404:
                # special case reporting for a 404
                raise UrlOperationsResourceUnknown(
                    url, status_code=e.response.status_code) from e
            else:
                raise UrlOperationsRemoteError(
                    url, message=str(e), status_code=e.response.status_code
                ) from e

    def stat(self,
             url: str,
             *,
             credential: str | None = None,
             timeout: float | None = None) -> Dict:
        """Gather information on a URL target, without downloading it

        See :meth:`datalad_next.url_operations.UrlOperations.stat`
        for parameter documentation and exception behavior.

        Raises
        ------
        UrlOperationsResourceUnknown
          For access targets found absent.
        """
        auth = DataladAuth(self.cfg, credential=credential)
        props: Dict[str, str | int]
        with requests.head(
                url,
                headers=self.get_headers(),
                auth=auth,
                # we want to match the `get` behavior explicitly
                # in order to arrive at the final URL after any
                # redirects that get would also end up with
                allow_redirects=True,
        ) as r:
            # fail visible for any non-OK outcome
            self._check_status(r, url)
            props = {
                # standardize on lower-case header keys.
                # also prefix anything other than 'content-length' to make
                # room for future standardizations
                k.lower() if k.lower() == 'content-length' else f'http-{k.lower()}':
                v
                for k, v in r.headers.items()
            }
            props['url'] = r.url
        auth.save_entered_credential(
            context=f"for accessing {url}"
        )
        if 'content-length' in props:
            # make an effort to return size in bytes as int
            try:
                props['content-length'] = int(props['content-length'])
            except (TypeError, ValueError):
                # but be reasonably robust against unexpected responses
                pass
        return props

    def delete(self,
               url: str,
               *,
               credential: str | None = None,
               timeout: float | None = None) -> Dict:
        """Delete the target of a http(s)://-URL

        """
        auth = DataladAuth(self.cfg, credential=credential)
        try:
            with requests.delete(
                    url=url,
                    stream=True,
                    headers=self.get_headers(),
                    auth=auth,
                    timeout=timeout,
            ) as r:
                # fail visible for any non-OK outcome
                self._check_status(r, url)
        except requests.exceptions.ReadTimeout as e:
            raise TimeoutError(f"Timeout while deleting {url}") from e
        auth.save_entered_credential(
            context=f'download from {url}'
        )
        return {}

    def download(self,
                 from_url: str,
                 to_path: Path | None,
                 *,
                 credential: str | None = None,
                 hash: list[str] | None = None,
                 timeout: float | None = None) -> Dict:
        """Download via HTTP GET request

        See :meth:`datalad_next.url_operations.UrlOperations.download`
        for parameter documentation and exception behavior.

        Raises
        ------
        UrlOperationsResourceUnknown
          For download targets found absent.
        """
        # a new manager per request
        # TODO optimize later to cache credentials per target
        # similar to requests_toolbelt.auth.handler.AuthHandler
        auth = DataladAuth(self.cfg, credential=credential)
        with requests.get(
                from_url,
                stream=True,
                headers=self.get_headers(),
                auth=auth,
        ) as r:
            # fail visible for any non-OK outcome
            self._check_status(r, from_url)
            download_props = self._stream_download_from_request(
                r, to_path, hash=hash)
        auth.save_entered_credential(
            context=f'download from {from_url}'
        )
        return download_props

    def probe_url(self, url, timeout=10.0, headers=None):
        """Probe a HTTP(S) URL for redirects and authentication needs

        This functions performs a HEAD request against the given URL,
        while waiting at most for the given timeout duration for
        a server response.

        Parameters
        ----------
        url: str
          URL to probe
        timeout: float, optional
          Maximum time to wait for a server response to the probe
        headers: dict, optional
          Any custom headers to use for the probe request. If none are
          provided, or the provided headers contain no 'user-agent'
          field, the default DataLad user agent is added automatically.

        Returns
        -------
        str or None, dict
          The first value is the URL against the final request was
          performed, after following any redirects and applying
          normalizations.

          The second value is a mapping with a particular set of
          properties inferred from probing the webserver. The following
          key-value pairs are supported:

          - 'is_redirect' (bool), True if any redirection occurred. This
            boolean property is a more accurate test than comparing
            input and output URL
          - 'status_code' (int), HTTP response code (of the final request
            in case of redirection).
          - 'auth' (dict), present if the final server response contained any
            'www-authenticate' headers, typically the case for 401 responses.
            The dict contains a mapping of server-reported authentication
            scheme names (e.g., 'basic', 'bearer') to their respective
            properties (dict). These can be any nature and number, depending
            on the respective authentication scheme. Most notably, they
            may contain a 'realm' property that can be used to determine
            suitable credentials for authentication.

        Raises
        ------
        requests.RequestException
          May raise any exception of the `requests` package, most notably
          `ConnectionError`, `Timeout`, `TooManyRedirects`, etc.
        """
        hdrs = self.get_headers()
        if headers is None:
            headers = hdrs
        elif 'user-agent' not in headers:
            headers.update(hdrs)

        props = {}
        req = requests.head(
            url,
            allow_redirects=True,
            timeout=timeout,
            headers=headers,
        )
        if 'www-authenticate' in req.headers:
            props['auth'] = parse_www_authenticate(
                req.headers['www-authenticate'])
        props['is_redirect'] = True if req.history else False
        props['status_code'] = req.status_code
        return req.url, props

    def _stream_download_from_request(
            self, r, to_path, hash: list[str] | None = None) -> Dict:
        from_url = r.url
        hasher = self._get_hasher(hash)
        progress_id = self._get_progress_id(from_url, to_path)
        # try to get download size, it might not be provided, e.g. if
        # chunked transport encoding is used
        try:
            # for compressed downloads the content length refers to the
            # compressed content
            expected_size = int(r.headers.get('content-length'))
        except (ValueError, TypeError):
            # some HTTP-200 responses do not have a `content-length` header,
            # e.g. if chunked transport encoding is used. in this case, set
            # up everything to calculate size by ourselves
            expected_size = None
        self._progress_report_start(
            progress_id,
            ('Download %s to %s', from_url, to_path),
            'downloading',
            # can be None, and that is OK
            expected_size,
        )

        fp = None
        props: Dict[str, str] = {}
        try:
            # we can only write to file-likes opened in bytes mode
            fp = sys.stdout.buffer if to_path is None else open(to_path, 'wb')
            # we need to track how much came down the pipe for progress
            # reporting
            downloaded_bytes = 0
            # TODO make chunksize a config item, 65536 is the default in
            # requests_toolbelt
            for chunk in r.raw.stream(amt=65536, decode_content=True):
                # update how much data was transferred from the remote server.
                if expected_size:
                    # if we have an expected size, we don't use the size of the
                    # chunk for that because content might be downloaded with
                    # transparent (de)compression. instead we ask the download
                    # stream itself for its "position".
                    tell = r.raw.tell()
                else:
                    # if we do not have an expected size, all we can use is
                    # the size of the downloaded chunk.
                    tell = downloaded_bytes + len(chunk)
                self._progress_report_update(
                    progress_id,
                    ('Downloaded chunk',),
                    tell - downloaded_bytes,
                )
                fp.write(chunk)
                downloaded_bytes = tell
                # compute hash simultaneously
                hasher.update(chunk)
            props.update(hasher.get_hexdigest())
            return props
        finally:
            if fp and to_path is not None:
                fp.close()
            self._progress_report_stop(progress_id, ('Finished download',))
